(*
  Bifton - A Pascal- compiler.

  Copyright (C) 2004 Samuel Mimram & Julien Cristau.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(**
  Abstract syntax tree with an assembly looking-like structure.

  @author Samuel Mimram
*)

(* $Id: asm_ir.ml,v 1.27 2004/05/09 19:32:58 smimou Exp $ *)

(** An error during the generation of the assembly. *)
exception Asm_error of string

(**
  Pseudo-typed values for assembly.
*)
type value =
  | Int of int
  | Bool of bool

(**
  Type of variables.
*)
type var_type =
  | Boolean
  | Integer
  | Pointer
  | Other of int (** arrays, etc (the argument is its size) *)
  | No_type (** no type was infered yet (this should not be generated) *)
(* | Float *)

(**
  Get the size in memory, in bytes, needed by a type.
*)
let sizeof = function
  | Boolean -> 4
  | Integer -> 4
  | Pointer -> 4
  | Other n -> n
  | No_type -> raise (Asm_error "internal error in sizeof (trying to get the size of No_type)")

(**
  Kinds of flags.
*)
type xflag =
  | Aflag (** above (unsigned) *)
  | Bflag (** below (unsigned) *)
  | Eflag (** equal *)
  | Zflag (** zero *)
  | Gflag (** greater than (signed) *)
  | Lflag (** less than (signed) *)
  | Cflag (** carry *)
  | Oflag (** overflow *)
  | Pflag (** parity *)
  | Sflag (** sign *)
  | Vflag of variable (** the result is contained in a variable *)

(**
  A flag.
*)
and flag =
    {
      flag_not : bool;
      flag_flag : xflag;
      flag_eq : bool;
    }

(**
  A register.
*)
and register =
  | Reg_int of int (** a register which can contain integers *)
  | Reg_flag of flag (** flags of the processor *)

(**
  A position on the stack.
*)
and stack_pos =
  | Pos_base of int (** position is relative to the base pointer *)
  | Pos_stack of int (** position is relative to the stack pointer *)
  | Pos_label of string (** position is a label in the asm *)
  | Pos_arr of variable * variable * int * int (** [Pos_arr(b, v, n0, s)] means [&b] + ([(v)] - [n0]) * [s] where [s] might be equal [1], [2], [4] or [8] (useful for arrays) *)
  | Pos_rec of variable * int (** [Pos_ref(b, k)] means [&b] + [k] (useful for records) *)
  | Pos_pointed of variable (** variable is pointed by an other variable *)
  | Pos_none (** no position has been assigned yet *)
  | Pos_reg of register (** used for function arguments on sparc *)

(**
  A variable definition.

  The [ref] are here because they should be shared between variable definitions.
*)
and variable_def =
    {
      mutable vd_ident : int; (** unique identifier *)
      mutable vd_pos : stack_pos ref; (** position, in bytes, relative to the frame pointer position *)
      mutable vd_type : var_type ref; (** type *)
      mutable vd_lex_depth : int ref; (** lexical depth *)
      mutable vd_value : value option; (** for constants *)
      mutable vd_on_the_stack: bool ref; (** should some space be allocated on the stack for the variable by default? (should be false for booleans and integers by default) *)
    }

(**
  A variable.
*)
and variable =
    {
      mutable var_reg : register option; (** number of its register *)
      mutable var_def : variable_def; (** position of the variable on the stack *)
    }

(**
  Get the negated version of a flag.
*)
let neg_flag f =
  {
    flag_not = not f.flag_not;
    flag_flag = f.flag_flag;
    flag_eq = f.flag_eq;
  }

(**
  Find the flag with arguments swapped (e.g. x < y ~ y > x).
*)
let commute_flag f =
  {
    flag_not = not f.flag_not;
    flag_flag = f.flag_flag
     (*
       match f.flag_flag with
	 | Aflag -> Bflag
	 | Bflag -> Aflag
	 | Eflag -> Eflag
	 | Lflag -> Gflag
	 | Gflag -> Lflag
     *);
    flag_eq = not f.flag_eq;
  }

(** Fresh identifiers generator. *)
let vd_counter = ref (-1)
(**
  [new_var_def type depth] generates a new variable definition.
*)
let new_var_def t lex_depth =
  {
    vd_ident = (incr vd_counter; !vd_counter);
    vd_pos = ref Pos_none;
    vd_type = ref t;
    vd_lex_depth = ref lex_depth;
    vd_value = None;
    vd_on_the_stack = ref false;
  }

(**
  Create a new variable.
*)
let new_var vd =
  {
    var_def = vd;
    var_reg = None;
  }

(**
  [new_const_int_var n depth] creates a new variable which is a constant integer equal to [n].
*)
let new_const_int_var n lex_depth =
  let vd = new_var_def Integer lex_depth in
    vd.vd_value <- Some (Int n);
    new_var vd

(**
  Get a copy of a variable definition.
*)
let rec copy_var_def vd =
  {
    vd_ident = vd.vd_ident;
    vd_type = vd.vd_type;
    vd_lex_depth = vd.vd_lex_depth;
    vd_value = vd.vd_value;
    vd_on_the_stack = vd.vd_on_the_stack;
    vd_pos = 
     match !(vd.vd_pos) with
       | Pos_pointed v -> ref (Pos_pointed (copy_var v))
       | Pos_rec(r, k) -> ref (Pos_rec (copy_var r, k))
       | Pos_arr(a, v, n0, s) -> ref (Pos_arr (copy_var a, copy_var v, n0, s))
       | Pos_base _
       | Pos_stack _
       | Pos_label _
       | Pos_reg _
       | Pos_none -> vd.vd_pos
  }

(**
  Get a copy of a variable.
*)
and copy_var v =
  {
    var_reg = v.var_reg;
    var_def = copy_var_def v.var_def;
  }

(**
  Get the register in which a variable is stored.
*)
let get_var_reg v =
  match v.var_reg with
    | Some r -> r
    | None -> raise Not_found (* raise (Asm_error "trying to get the register of a variable which is not in a register") *)

let set_var_reg v r =
  v.var_reg <- Some r

let get_vd_type vd = !(vd.vd_type)

let get_var_type v = get_vd_type v.var_def

(**
  Get the lexical depth of a variable definition.
*)
let get_vd_lex vd = !(vd.vd_lex_depth)

(**
  Get the lexical depth of a variable.
*)
let get_var_lex v = get_vd_lex v.var_def

(**
  Get the position of a variable definition on the stack.
*)
(* TODO: s/var_def/vd/ *)
let get_vd_pos vd = !(vd.vd_pos)

(**
  Get the position of a variable on the stack.
*)
let get_var_pos v = get_vd_pos v.var_def

(**
  Is a variable definition on the stack?
*)
let get_vd_ots vd = !(vd.vd_on_the_stack)

(**
  Is a variable on the stack?
*)
let get_var_ots v = get_vd_ots v.var_def

let set_vd_ots vd = vd.vd_on_the_stack := true

(* TODO: use this *)
let set_var_ots v = set_vd_ots v.var_def

let get_vd_value vd = vd.vd_value

let get_var_value var = get_vd_value var.var_def

(**
  A procedure.
*)
type procedure =
    {
      proc_name : string; (** name of the procedure *)
      proc_args : variable_def list; (** arguments of the procedure *)
      mutable proc_vars : variable_def list; (** variables defined in the procedure (including temporary ones). *)
      mutable proc_value : base_block; (** first base block of the procedure *)
      mutable proc_lex_depth : int; (** lexical depth *)
      mutable proc_stack_size : int; (** initial size of the stack *)
      mutable proc_is_fun : bool (** is it a function? *)
    }

(**
  A program.
*)
and program =
    {
      mutable prog_source_name : string; (** filename of the source code *)
      prog_procs : procedure list; (** procedures of the program (must be a prefix in-depth "parcours" of the tree) *)
      prog_max_depth : int; (** maximal lexical depth *)
    }

(**
  A meta instruction (an instruction which has several base blocks as argument).
*)
and meta_instr =
  | If of variable * base_block * base_block * base_block (** a if ... then ... else ... construction *)
  | While of variable * base_block * base_block * base_block (** a while do ... done construction (the first base block is here to compute the condition) *)
  | End

(**
  An assembly instruction.
*)
and instr =
    (* TODO: add other operators *)
  | Add_i of variable * variable * variable (** x1 <- x2 + x3 where the xi are integers *)
  | And of variable * variable * variable (** x1 <- x2 && x3 *)
  | Cmp_i of variable * variable * flag * variable (** compare x2 and x3 where the xi are integers and put the result in x1 *)
  | Delete of variable (** free the memory used by a variable *)
  | Div_i of variable * variable * variable (** x1 <- x2 / x3 where the xi are integers *)
  | Mod_i of variable * variable * variable (** x1 <- x2 mod x3 where the xi are integers *)
  | Mov_i of variable * variable (** x1 <- x2 where the xi are integers *)
  | Mov_b of variable * variable (** x1 <- x2 where the xi are booleans *)
  | Mult_i of variable * variable * variable (** x1 <- x2 * x3 where the xi are integers *)
  | Neg_i of variable * variable (** x1 <- -x2 *)
  | New of variable * int (** allocate a value of which size is specified in bytes *)
  | Not of variable * variable (** x1 <- not x2 *)
  | Or of variable * variable * variable (** x1 <- x2 && x3 *)
  | Sub_i of variable * variable * variable (** x1 <- x2 - x3 *)
  | Fun of variable * procedure * variable list (** x <- f(args) *)
  | Return of variable (** return a value (for functions) *)
  | Proc of procedure * variable list (** a procedure call *)
  | Nop (* TODO: remove *) (** don't do anything *)
  | Read_flag of flag * variable (** set a variable's contents according to a flag *)
  | Address_of of variable * variable (** x1 <- &x2 *)
  | Spill_var of variable

  (** The following are not generated but can be used in code generation. *)
  | Assert_avail of variable (** make sure the variable is available in a register after this instuction *)
  | Check_avail of variable (** we are going to use this variable, make sure that everything is going to be ok *)
  | Check_avail_with_flags of variable (** like [Check_avail] but the variable might be stored in flags *)
  | Call of string (** a call *)
  | Push of variable (** push a variable on the top of the stack *)
  | Push_o of variable * int (** push [n] bytes on the stack *)
  | Spill of bool * register * variable_def (** spill a variable (store it on the stack); the first argument should be set to [true] iff we are not in the same lexical scope as the variable *) (* TODO: do we need both arguments? *)

  | Unspill of bool * variable (** unspill a variable (cf. [Spill] for the first argument) *)
  | Spill_all (** spill all variables still in a register *)
  | Set_var of flag * variable (** set a register to 0 or 1 according to some flags *)
  | Code of string (** some verbatim code (should only be used for debugging purposes) *) (* TODO: this is ugly, remove *)

(**
  A base block.
*)
(* TODO: add fields for data-flows *)
and base_block =
    {
      mutable bb_block : instr list;
      mutable bb_link : meta_instr;
      (* bb_used_vars : variable_def list; *)
    }

(**
  Create a new base block.
*)
let new_base_block mi =
  {
    bb_block = [];
    bb_link = mi;
  }

(**
  Create a stub for a procedure, given its name, arguments and lexical depth.
*)
let proc_stub name args lex_depth =
  {
    proc_name = name;
    proc_args = args;
    proc_vars = [];
    proc_value = new_base_block End;
    proc_lex_depth = lex_depth;
    proc_stack_size = 0;
    proc_is_fun = false;
  }

(**
  Is a variable a constant?
*)
let is_constant v =
  match v.var_def.vd_value with
    | Some _ -> true
    | None -> false

(**
  Use commutativity on an instruction.
*)
let commute_instr = function
  | Add_i(v1, v2, v3) -> Add_i(v1, v3, v2)
  | And(v1, v2, v3) -> And(v1, v3, v2)
  | Mult_i(v1, v2, v3) -> Mult_i(v1, v3, v2)
  | Or(v1, v2, v3) -> Or(v1, v3, v2)
  | Cmp_i(v1, v2, f, v3) -> Cmp_i(v1, v3, neg_flag f, v2)
  | Not _
  | Code _
  | Set_var _
  | Spill _
  | Unspill _
  | Push_o _
  | Push _
  | Call _
  | Check_avail _
  | Read_flag _
  | Check_avail_with_flags _
  | Assert_avail _
  | Spill_var _
  | Address_of _
  | Proc _
  | Return _
  | Fun _
  | Sub_i _
  | Neg_i _
  | Mov_b _
  | Mov_i _
  | Mod_i _
  | Div_i _
  | New _
  | Delete _
  | Spill_all
  | Nop -> raise (Asm_error "internal error in commute_instr (operation does not commute)")
