(*
  Bifton - A Pascal- compiler.

  Copyright (C) 2004 Samuel Mimram & Julien Cristau.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(**
  i386 backend.

  In 3-addresses codes, the first one is not used and the second will contain the result.
  %edi is reserved for locally needed registers.

  Push_o needs %ecx, %esi and %edi.

  @author Samuel Mimram
*)

(* $Id: i386.ml,v 1.34 2004/05/09 19:32:58 smimou Exp $ *)

open Asm_ir

let string_of_char c =
  String.make 1 c

let rec list_assoc_id x = function
  | (a, b) :: t when x.vd_ident = a.vd_ident -> b
  | _ :: t -> list_assoc_id x t
  | [] -> raise Not_found

(* TODO: improve *)
let rec list_mem_assoc_id x l =
  try
    ignore (list_assoc_id x l);
    true
  with
    | Not_found -> false

let rec list_assoc' x = function
  | (a, b) :: t when x = b -> a
  | _ :: t -> list_assoc' x t
  | [] -> raise Not_found

let rec list_mem_assoc' x = function
  | (a, b) :: t when x = b -> true
  | _ :: t -> list_mem_assoc' x t
  | [] -> false

let list_itern f l =
  let rec aux n = function
    | h :: t -> f n h; aux (n + 1) t
    | [] -> ()
  in
    aux 0 l

let esp = Reg_int (-2)
let ebp = Reg_int (-1)
let eax = Reg_int 0
let ebx = Reg_int 1
let ecx = Reg_int 2
let edx = Reg_int 3
let esi = Reg_int 4
let edi = Reg_int 5

(**
  Get a as-like string representation of a register.
*)
let string_of_register = function
  | Reg_int (-2) -> "%esp"
  | Reg_int (-1) -> "%ebp"
  | Reg_int 0 -> "%eax"
  | Reg_int 1 -> "%ebx"
  | Reg_int 2 -> "%ecx"
  | Reg_int 3 -> "%edx"
  | Reg_int 4 -> "%esi"
  | Reg_int 5 -> "%edi"
  | Reg_flag _ -> raise (Asm_error "trying to use flags as register")
  | _ -> raise (Asm_error "trying to use a register > %edx!")

(**
  Get a variable allocated in a specific register.
*)
let get_reg_var r =
  let vd = new_var_def Pointer (-1) in
  let v = new_var vd in
    v.var_reg <- Some r; v

(**
  Get a variable containing the address of the top of the stack.
*)
let get_stack_pos_var () =
  get_reg_var esp

(** Total number of available registers (warning: register numbers are zero-based). *)
let nb_registers = 5

(**
  Do the registers allocation. This is done base block by base block (all registers should be available at the beginning of each base block and all variables needed again should be spilled at the end of each base block). %edx is reserved for loading values passed by reference.
*)
let allocate_registers base_block lex_depth =
  (**
    @param av list of available variables (list of definition and associated register)
    @param instr the instruction code should be generated for
    @param needed_defs definition of variables whose values are used in the rest of the block
    @returns a list of available variables after the instruction and a rewrite of the code *)
  (* TODO: really compute and use needed_defs *)
  let rec alloc_instr av instr needed_defs =
    let av = ref av in
      (** Declare that a variable is avaiable in a register. If some other variable was previously declared as available in the same register, it won't be anymore. *)
    let set_available v r =
      if list_mem_assoc_id v !av then failwith "a variable has two locations in memory!";
      av := (v, r) :: (List.filter (fun (_, r') -> r' <> r) !av)
    in

    (** Warning [before] and [after] are reversed at the end of the function! *)
    let before = ref []
    and after = ref [] in (* TODO: remove after if it's not used *)
    let do_before a =
      before := a :: !before
    in
    let do_after a =
      after := a :: !after
    in
      
    let during =
      (** Mark a register as unused. *)
      let free_reg r =
	av := List.filter (fun (_, r') -> r' <> r) !av
      in

      (** Spill a variable. *)
      let rec do_spill r vd =
	vd.vd_on_the_stack := true;
	match get_vd_type vd with
	  | Other _ -> ()
	  | _ -> do_before (Spill(get_vd_lex vd <> lex_depth, r, vd))

      (** Unspill a variable. *)
      and do_unspill v =
	match get_var_type v with
	  | Other _ ->
	      do_before (Address_of(v, v)) (* TODO: this is not nice at all + we should determine if we're in the same lexical scope or not *)
	  | _ ->
	      do_before (Unspill (get_var_lex v <> lex_depth, v))

      (** Allocate a variable in a specific register. *)
      and alloc_in_reg v r occ =
	(** Warning: v.var_reg should be set for flags! *)
	match v.var_reg with
	  | Some ((Reg_flag f) as rf) ->
	      v.var_reg <- Some r;
	      do_before (Set_var(f, v));
	      free_reg rf;
	      set_available v.var_def r; r
	  | Some r -> r (* TODO: do not compute the first r in this case! *)
	  | None ->
	      (
		match v.var_def.vd_value with
		  | Some _ ->
		      v.var_reg <- Some r;
		      let v' = copy_var v in
			v'.var_reg <- None;
			do_before (Mov_i(v, v'));
			set_available v.var_def r; r
		  | None ->
		      v.var_reg <- Some r;
		      do_unspill (copy_var v);
		      set_available v.var_def r; r
	      )
		
      (** Allocate an integer register to the variable [v].
	@param occ is a list of already taken integer registers *)
      and alloc_reg v occ =
	(* TODO: should this be there or in check_avail? *)
	let occ =
	  match get_var_pos v with
	    | Pos_arr(a, v, _, _) -> get_var_reg a :: get_var_reg v :: occ
	    | Pos_rec(p, _)
	    | Pos_pointed p -> get_var_reg p :: occ
	    | Pos_base _
	    | Pos_stack _
	    | Pos_label _
	    | Pos_reg _
	    | Pos_none -> occ
	in
	let r =

	  (** Find an unused register. *)
	  let rec find_free_int n =
	    if n >= nb_registers then raise Not_found
	    else if List.mem (Reg_int n) occ then
	      find_free_int (n + 1)
	    else if list_mem_assoc' (Reg_int n) !av then
	      find_free_int (n + 1)
	    else
	      Reg_int n
	  in
	    (* TODO: spill the one which will not be used during the most long time *)
	  let rec find_to_spill_int n =
	    if n >= nb_registers then raise (Asm_error "no registers left during assembly generation")
	    else if List.mem (Reg_int n) occ then
	      find_to_spill_int (n + 1)
	    else
	      Reg_int n
	  in
	    try
	      find_free_int 0
	    with
	      | Not_found ->
		  let r = find_to_spill_int 0 in
		  let replaced_vd = list_assoc' r !av in
		    spill_free true r replaced_vd occ; r
	in
	  alloc_in_reg v r occ
	    
      (* TODO: if it's not needed anymore, don't spill it *)
      (** @param free should we declare the freed variable's register as free?
	@param occ is here to specify needed registers because we might need an extra register during the spill *)
      and spill_free free r vd occ =
	vd.vd_on_the_stack := true;
	(
	  match get_vd_pos vd with
	    | Pos_arr(b, v, _, _) ->
		assert_avail b occ; assert_avail v occ
	    | Pos_pointed p
	    | Pos_rec(p, _) -> assert_avail p occ
	    | _ -> ()
	);
	(
	  match r with
	    | Reg_int _ ->
		do_spill r vd;
		if free then free_reg r
	    | Reg_flag f ->
		let v = new_var vd in
		  v.var_reg <- Some r;
		  free_reg r; (* TODO: this is not really nice but else we get the "two memory locations" error *)
		  (
		    (* TODO: why do we need to do that? *)
		    try
		      v.var_reg <- Some (list_assoc_id v.var_def !av)
		    with
		      | _ -> ignore (alloc_reg v occ)
		  );
		  let r' = get_var_reg v in
		    do_spill r' vd;
		    free_reg r';
		    if free then free_reg r
	);

      and spill_var_free free v occ =
	check_avail_flags true v occ (fun l -> List.iter (fun r -> spill_free free r v.var_def occ) l)

      (** Make sure that a specific register is available. *)
      and spill_reg r occ =
	List.iter (fun (vd, rd) -> if r = rd then spill_free true rd vd occ) !av
	  
      (** If the variable [v] is available then we should execute [f] which will receive [v]'s register as argument else we should execute [g]. *)
      and on_avail v f g =
	try
	  let r = list_assoc_id v.var_def !av in
	    f r
	with
	  | Not_found -> g ()

      (** If variable [v] already was in a register it stays in it. The function [f] is called with the list of used registers. The second argument tells whether it is acceptable or not for it to be stored in flags. *)
      and check_avail_flags flags v occ f =
	let check_pos occ = function
	  | Pos_rec(v, _)
	  | Pos_pointed v -> assert_avail v occ; get_var_reg v :: occ
	  | Pos_arr(a, v, _, _) ->
	      assert_avail a occ;
	      let occ = get_var_reg a :: occ in
		assert_avail v occ;
		get_var_reg v :: occ
	  | Pos_base _ | Pos_stack _ | Pos_label _ | Pos_reg _ | Pos_none -> occ
	in
	  v.var_reg <- None; (* TODO: is that useful? *)
	  on_avail v
	    (fun r ->
	       (* let occ = check_pos (r :: occ) v.var_def.vd_pos in *)
	       let r =
		 if flags && !Params.use_flags then r
		 else
		   match r with
		     | Reg_flag _ ->
			 v.var_reg <- Some r;
			 alloc_reg v occ
		     | _ -> r
	       in
		 v.var_reg <- Some r; f [r]
	    )
	    (fun () ->
	       let occ = check_pos occ (get_var_pos v) in
		 if get_var_lex v <> lex_depth || get_var_type v = Boolean then
		   f [alloc_reg v occ]
		 else
		   (
		     if not (is_constant v) then
		       v.var_def.vd_on_the_stack := true;
		     f []
		   )
	    )

      (** If the variable is not yet in a register put it in some register. *)
      and assert_avail v occ =
	check_avail_flags false v occ
	  (fun r ->
	     match r with
	       | [] -> ignore (alloc_reg v occ)
	       | [Reg_flag f] -> ignore (alloc_reg v occ) 
	       | _ -> ()
	  )
      in
      let spill = spill_free true in

      (** If the variable is in a register then spill it. Its register is then marked as free. *)
      let spill_var = spill_var_free true in
	
      let spill_all () =
	List.iter (fun (vd, r) ->
		     match r with
		       | Reg_int _ ->
			   vd.vd_on_the_stack := true;
			   spill r vd []
		       | Reg_flag _ -> () (* TODO! *)
		  ) !av
      in
      let check_avail = check_avail_flags false in
      let just_check_avail v occ =
	check_avail v occ (fun _ -> ())
      in
      let used_and_replaced v occ =
	if !(Params.minimize_spills) then
	  (
	    spill_var_free false v occ;
	    assert_avail v occ; (* TODO: we do not need to load the value *)
	    free_reg (get_var_reg v)
	  )
	else
	  (
	    spill_var v occ;
	    assert_avail v occ
	  )
      in
	(* TODO: do not systematically free with spill_var! *)
	match instr with
	  | Mov_b(v1, v2)
	  | Mov_i(v1, v2)
	  | Address_of(v1, v2) ->
	      check_avail v2 []
		(fun r2 ->
		   spill_var v1 r2;
		   (* TODO: we don't need to load v1's value *)
		   assert_avail v1 r2;
		   (* used_and_replaced v1 r2; *)
		   (* set_available v1.var_def (get_var_reg v1) *)
		);
	      (* check_avail v2 []
		 (fun r2 -> ignore (alloc_reg v1 r2)); *)
	      (
		match get_var_pos v1 with
		  | Pos_pointed _
                  | Pos_arr _
                  | Pos_rec _ -> do_after (Spill(false, get_var_reg v1, v1.var_def)); free_reg (get_var_reg v1)
                  | _ -> ()
              );
	      [instr]
	  | New(v1, n) ->
	      used_and_replaced v1 [];
	      set_available v1.var_def eax;
	      [instr]

	  (** Not commutative operations. *)
	  (** [v2] is going to be replaced. *)
	  | Sub_i(v1, v2, v3) ->
	      check_avail v3 []
		(fun r3 ->
		   spill_var_free false v2 r3;
		   set_available v1.var_def (alloc_reg v2 r3)
		);
	      [instr]
	  | Not(v1, v2)
	  | Neg_i(v1, v2) ->
	      used_and_replaced v2 [];
	      let r2 = get_var_reg v2 in
		set_available v1.var_def r2;
		[instr]
	  | Cmp_i(v1, v2, flags, v3) ->
	      if is_constant v2 then
		(
		  assert (not (is_constant v3)); (* TODO *)
		  let av', i = alloc_instr !av (Cmp_i(v1, v3, commute_flag flags, v2)) needed_defs in
		    av := av'; i
		)
	      else
		(
		  (** If there was already somebody in the flags we spill it. *)
		  List.iter
		   (fun (vd, r) ->
		      match r with
			| Reg_flag f -> spill r vd []
			| _ -> ()
		   ) !av;
		  check_avail v3 []
		    (fun r3 ->
		       assert_avail v2 r3;
		       let r2 = get_var_reg v2 in
			 set_available v1.var_def (Reg_flag flags)
		    );
		  [instr]
		)

	  (** Commutative operations. *)
	  | And(v1, v2, v3)
	  | Or(v1, v2, v3)
	  | Add_i(v1, v2, v3) ->
	      if is_constant v2 then
		(
		  assert (not (is_constant v3));
		  let av', i = (alloc_instr !av (commute_instr instr) needed_defs) in
		    av := av'; i
		)
	      else
		(
		  check_avail v3 []
		   (fun r3 ->
		      used_and_replaced v2 r3;
		      let r2 = get_var_reg v2 in
			set_available v1.var_def r2
		   );
		  [instr]
		)
	  | Mult_i(v1, v2, v3) ->
	      if is_constant v2 then
		(
		  assert (not (is_constant v3));
		  let av', i = (alloc_instr !av (commute_instr instr) needed_defs) in
		    av := av'; i
		)
	      else
		(
		  (* TODO: use commutativity *)
		  (** I386 is so dirty... The result is stored in %edx:%eax. *)
		  let occ = [eax; edx] in
		    List.iter (fun r -> spill_reg r occ) occ;
		    check_avail v3 occ
		      (fun r3 ->
			 spill_var v2 (r3 @ occ);
			 ignore (alloc_in_reg v2 eax (r3 @ occ));
			 free_reg eax;
			 free_reg edx;
			 set_available v1.var_def eax
		      );
		    [instr]
		)
	  | Mod_i(v1, v2, v3)
	  | Div_i(v1, v2, v3) ->
	      (** I386 is so dirty... The result is stored in %edx:%eax. *)
	      let occ = [eax; edx] in
		List.iter (fun r -> spill_reg r occ) occ;
		assert_avail v3 occ;
		let r3 = [get_var_reg v3] in
		  spill_var v2 (r3 @ occ);
		  ignore (alloc_in_reg v2 eax (r3 @ occ));
		  free_reg eax;
		  free_reg edx;
		  set_available v1.var_def
		    (Reg_int
		       (
			 match instr with
			   | Div_i _ -> 0
			   | Mod_i _ -> 3
			   | _ -> raise (Asm_error "internal error")
		       )
		    );
		  [instr]
	  | Proc(p, l) ->
	      (** Let's save used registers. *)
	      spill_all ();
	      (** Those registers might be needed for copying arrays and records. *)
	      let occ = [ecx; esi; edi] in
	      let pushes = 
		(List.rev
		   (List.fold_left2
		      (fun l a v ->
			 (
			   just_check_avail v occ;
			   match get_var_type v with
			     | Other _ ->
				 assert_avail v occ;
				 Push_o(v, sizeof (get_var_type v))
			     | _ ->
				 Push v
			 ) :: l
		      ) [] p.proc_args l))
	      in
		av := [];
		pushes @ [Call p.proc_name;
			  Add_i (get_stack_pos_var (),
				 get_stack_pos_var (),
				 new_const_int_var (List.fold_left (fun s a -> s + (sizeof (get_vd_type a))) 0 p.proc_args) lex_depth)]
	  | Fun(x, p, l) ->
	      (** Let's save used registers. *)
	      spill_all ();
	      (** Those registers might be needed for copying arrays and records. *)
	      let occ = [ecx; esi; edi] in
	      let pushes = 
		(List.rev
		   (List.fold_left2
		      (fun l a v ->
			 (
			   just_check_avail v occ;
			   match get_var_type v with
			     | Other _ ->
				 assert_avail v occ;
				 Push_o(v, sizeof (get_var_type v))
			     | _ ->
				 Push v
			 ) :: l
		      ) [] p.proc_args l))
	      in
		av := [];
		used_and_replaced x occ;
		set_available x.var_def eax;
		pushes @ [Call p.proc_name;
			  Add_i (get_stack_pos_var (),
				 get_stack_pos_var (),
				 new_const_int_var (List.fold_left (fun s a -> s + (sizeof (get_vd_type a))) 0 p.proc_args) lex_depth)]
	  | Delete v
	  | Return v ->
	      just_check_avail v [];
	      [instr]
	  | Nop -> []
	  | Assert_avail v -> assert_avail v []; []
	  | Check_avail_with_flags v -> check_avail_flags false v [] (fun _ -> ()); []
	  | Spill_var v ->
	      spill_var v []; []
	  | Spill_all ->
	      spill_all (); []
	  | Code _
	  | Set_var _
	  | Unspill _
	  | Spill _
	  | Push_o _
	  | Push _
	  | Call _
	  | Check_avail _
	  | Read_flag _ -> raise (Asm_error "internal error in alloc_instr (encountered an instruction which should not have been generated)")
    in
      !av, (List.rev !before) @ during @ (List.rev !after)
  in
    (* TODO: data-flow equations *)
    (*
      let get_needed_var_defs_in_instr = function
      | Mov_i(_, v2)
      | Mov_b(_, v2) -> [v2]
      | Add_i(_, v2, v3)
      | Cmp_i(_, v2, _, v3)
      | Mult_i(_, v2, v3) -> [v2; v3]
      in
    *)
  let rec get_needed_var_defs i =
    (* TODO: remove duplicates *)
    (* List.concat (List.map get_needed_var_defs_in_instr i) *)
    []
  in
  let used_var_defs_in_instrs = get_needed_var_defs
  in

  (** Allocate registers for a list of instructions.
    @param nd list of variable definition needed after the block
    @returns a list of available variables after the instruction and a rewrite of the code *)
  (* TODO: fill nd correctly *)
  let rec alloc_instrs av nd = function
    | h :: t ->
	let av', i' = alloc_instr av h [] (*used_var_defs_in_instrs t*) in (* TODO: use nd *)
	let av'', i'' = alloc_instrs av' nd t in
	  av'', i' @ i''
    | [] -> av, []
  in

  (**
    Allocate registers for the base block.
  *)
  let rec alloc_base_block av base_block =
    (** Assembly needed by the following base blocks. *)
    let needed =
      match base_block.bb_link with
	| If(v, _, _, _) -> [Check_avail_with_flags v]
	| While _
	| End -> []
    in
      (* TODO: the Spill_all is not really nice *)
      (* TODO: compute variables needed in the next blocks *)
    let av, instrs = alloc_instrs av [] (base_block.bb_block @ [Spill_all] @ needed) in
    let instrs = ref instrs in
      (
	match base_block.bb_link with
	  | If(v, b1, b2, b3) ->
	      let av1 = alloc_base_block av b1 in
	      let av2 = alloc_base_block av b2 in
		ignore (alloc_base_block
			  [] (* TODO: we could use the "intersection" of av1 and av2 *)
			  b3)
	  | While(v, be, b1, b2) ->
	      be.bb_block <- be.bb_block @ [Check_avail_with_flags v];
	      let ave = alloc_base_block av be in
	      let av1 = alloc_base_block ave b1 in (* TODO: is ave correct? *)
		ignore (alloc_base_block av1 b2)
	  | End -> ()
      );
      base_block.bb_block <- !instrs; av
  in
    ignore (alloc_base_block [] base_block)


(**
  Do some local opimizations on a base block.
*)
let rec local_opt_base_block base_block =
  let opt_instr2 x y =
    match (x, y) with
      | Unspill(_, v1), Mov_i(v1', _)
      | Mov_i(v1, _), Mov_i(v1', _) when get_var_reg v1' = get_var_reg v1 -> None
      | _ -> Some x
  in
    (* TODO: add some more *)
  let opt_instr = function
    | Add_i(v1, v2, v3) when get_var_value v3 = Some (Int 0) -> None
    | Div_i(v1, v2, v3) when get_var_value v3 = Some (Int 1) -> None
    | Div_i(v1, v2, v3) when get_var_value v2 = Some (Int 0) -> None
    | Sub_i(v1, v2, v3) when get_var_value v3 = Some (Int 0) -> None
    | x -> Some x
  in
  let rec opt_instrs = function
    | x :: y :: t ->
	(
	  match opt_instr2 x y with
	    | None -> opt_instrs (y :: t)
	    | Some x' when x' == x -> (opt_instrs [x']) @ (opt_instrs (y :: t))
	    | Some x' -> opt_instrs (x' :: y :: t)
	);
    | [x] ->
	(
	  match opt_instr x with
	    | Some x' -> [x']
	    | None -> []
	)
    | [] -> []
  in
    (
      match base_block.bb_link with
	| If(_, b1, b2, b3) ->
	    local_opt_base_block b1;
	    local_opt_base_block b2;
	    local_opt_base_block b3
	| While(_, be, b1, b2) ->
	    local_opt_base_block be;
	    local_opt_base_block b1;
	    local_opt_base_block b2
	| End -> ()
    );
    base_block.bb_block <- opt_instrs base_block.bb_block


(**
  Generate the target asm.
*)
let generate_code prog =
  let ret = ref "" in
  let next_label = ref 0 in
  let get_fresh_label () =
    incr next_label;
    "L" ^ (string_of_int !next_label)
  in
  let code_of_proc p =
    let string_of_flag f =
      (if f.flag_not then "n" else "") ^
      (match f.flag_flag with
	 | Aflag -> "a"
	 | Bflag -> "b"
	 | Cflag -> "c"
	 | Eflag -> "e"
	 | Gflag -> "g"
	 | Lflag -> "l"
	 | Oflag -> "o"
	 | Pflag -> "p"
	 | Sflag -> "s"
	 | Zflag -> "z"
	 | Vflag _ -> raise (Asm_error "internal error in string_of_flag")
      ) ^
      (if f.flag_eq then "e" else "")
    in
    let jump_of_flag f =
      "j" ^ (string_of_flag f)
    in

    (** Generate the code for accessing to the ARP of an other procedure. *)
    let get_display lex_depth r =
      "\tmovl\t(DISPLAY" ^ (if lex_depth = 0 then "" else "+" ^ string_of_int (4 * lex_depth)) ^ "), " ^ (string_of_register r) ^ "\n"
    in
    let code_of_instr instr =
      let before = ref "" in
      let do_before s =
	before := !before ^ s
      in

      (** The the as-like representation of a position on the stack.
	@param base the base pointer (might not be ebx because of lexical scopes) *)
      (* TODO: check that base is correctly used *)
      let rec string_of_stack_pos base pos =
	(** Get the register and the offset. *)
	let rec get_reg_n = function
	  | Pos_base n -> (* "%ebp" *)
	      string_of_register base, n
	  | Pos_stack n -> "%esp", n  
	  | Pos_pointed v -> code_of_var v, 0
	  | Pos_rec _
	  | Pos_arr _
	  | Pos_label _
	  | Pos_reg _
	  | Pos_none -> raise (Asm_error "internal error in get_reg_n")
	in
	  match pos with
	    | Pos_label n -> "$" ^ n
	    | Pos_rec(r, k) ->
		(string_of_int k) ^ "(" ^ string_of_register (get_var_reg r) ^ ")"
	    | Pos_arr(a, v, n0, s) ->
		(
		  let ar, an = string_of_register (get_var_reg a), 0 in
		    match v.var_def.vd_value with
		      | Some (Int i) ->
			  (string_of_int ((i - n0) * s)) ^ "(" ^ ar ^ ")"
		      | _ ->
			  let vr = get_var_reg v in
			    if s = 1 || s = 2 || s = 4 || s = 8 then
			      (string_of_int (an - n0 * s)) ^ "(" ^ ar ^ "," ^ string_of_register vr ^ "," ^ (string_of_int s) ^ ")"
			    else
			      (
				do_before
				 (
				   "\txor\t%edi, %edi\n" ^
				   "\tpushl\t%ebp\n" ^
				   (
				     let ans = ref "" in
				     let s = ref s in
				       for i = 0 to 31
				       do
					 if !s mod 2 = 1 then
					   ans := !ans ^
					   "\tmovl\t" ^ (string_of_register vr) ^ ", %ebp\n" ^
					   "\tsall\t$" ^ (string_of_int i) ^ ", %ebp\n" ^
					   "\taddl\t%ebp, %edi\n";
					 s := !s lsr 1
				       done;
				       !ans
				   ) ^
				   "\tpopl\t%ebp\n"
				 );
				(string_of_int (an - n0 * s)) ^ "(" ^ ar ^ ",%edi,1)"
				  
			      )
		)
	    | Pos_none -> raise (Asm_error "trying to get the position of variable which is not on the stack")
	    | Pos_base _
	    | Pos_stack _
	    | Pos_reg _
	    | Pos_pointed _ ->
		let r, n = get_reg_n pos in
		  (if n = 0 then "" else string_of_int n) ^ "(" ^ r ^ ")"

      (** @return register's name (or maybe a constant value). *)
      and code_of_var_no_reg v =
	match v.var_def.vd_value with
	  | Some(Int i) -> "$" ^ (string_of_int i)
	  | Some(Bool false) -> "$0"
	  | Some(Bool true) -> "$1"
	  | _ -> string_of_stack_pos ebp (get_var_pos v)
      and code_of_var_reg v =
	match v.var_reg with
	  | Some r -> string_of_register r
	  | None -> code_of_var_no_reg v
      and code_of_var v =
	if is_constant v then
	  code_of_var_no_reg v
	else
	  code_of_var_reg v
      in
	!before ^
	match instr with
	  | Add_i(v1, v2, v3) ->
	      let r2 = code_of_var_reg v2 in
	      let r3 = code_of_var v3 in
		"\taddl\t" ^ r3 ^ ", " ^ r2 ^ "\n"
	  | Mult_i(v1, v2, v3) ->
	      let r2 = code_of_var_reg v2 in
	      let r3 = code_of_var v3 in
		"\timull\t" ^ r3 ^ ", " ^ r2 ^ "\n"
		  
	  (** [v2] is supposed to be in %eax. *)
	  | Mod_i(v1, v2, v3)
	  | Div_i(v1, v2, v3) ->
	      let r2 = code_of_var v2 in
	      let r3 = code_of_var_reg v3 in
		"\tcltd\n" ^ "\tidivl\t" ^ r3 ^ "\n"
	  | Cmp_i(v1, v2, _, v3) ->
	      let r2 = code_of_var v2 in
	      let r3 = code_of_var v3 in
		"\tcmpl\t" ^ r3 ^ ", " ^ r2 ^ "\n"
	  | Sub_i(v1, v2, v3) ->
	      let r2 = code_of_var_reg v2 in
	      let r3 = code_of_var v3 in
		"\tsubl\t" ^ r3 ^ ", " ^ r2 ^ "\n"
	  | Not(v1, v2)
	  | Neg_i(v1, v2) ->
	      let r2 = code_of_var v2 in
		"\tnegl\t" ^ r2 ^ "\n"
	  | And(v1, v2, v3) ->
	      let r2 = code_of_var_reg v2 in
	      let r3 = code_of_var v3 in
		"\tandl\t" ^ r3 ^ ", " ^ r2 ^ "\n"
	  | Or(v1, v2, v3) ->
	      let r2 = code_of_var_reg v2 in
	      let r3 = code_of_var v3 in
		"\torl\t" ^ r3 ^ ", " ^ r2 ^ "\n"
	  | Mov_b(v1, v2)
	  | Mov_i(v1, v2) ->
	      if v2.var_def.vd_value = Some(Int 0) then
		let r1 = code_of_var_reg v1 in
		  "\txor\t" ^ r1 ^ ", " ^ r1 ^ "\n"
	      else
		(
		  let ans =
		    let r1 = code_of_var_reg v1 in
		    let r2 = code_of_var v2 in
		      "\tmovl\t" ^ r2 ^ ", " ^ r1 ^ "\n"
		  in
		    (** If both registers are the same then we can do nothing. *)
		    match v1.var_reg with
		      | Some r1 ->
			  (
			    match v2.var_reg with
			      | Some r2 when r2 = r1 -> ""
			      | _ -> ans
			  )
		      | _ -> ans
		)
	  | New(v1, n) ->
	      "\tpushl\t$" ^ string_of_int n ^ "\n" ^
	      "\tcall\tsbrk\n"
	  | Delete v1 ->
	      ""
	  | Address_of(v1, v2) ->
	      let r1 = code_of_var_reg v1 in
		(* TODO: check if we're in the same lexical scope *)
		(* if get_var_lex v1 <> get_var_lex v2 then *)
		get_display (get_var_lex v2) (get_var_reg v1) ^
		"\tleal\t" ^ (string_of_stack_pos (get_var_reg v1) (get_var_pos v2)) ^ ", " ^ r1 ^ "\n"
		  (* else
		     "\tleal\t" ^ (code_of_var_no_reg v2) ^ ", " ^ r1 ^ "\n" *)
	  | Unspill(lex_scope, v) ->
	      if not (get_var_ots v) then "" (* TODO: raise an error? *)
	      else
		(
		  let r' = code_of_var v in
		    if lex_scope then
		      match get_var_pos v with
			| Pos_base b ->
			    (
			      match v.var_reg with
				| None -> raise (Asm_error "trying to access a variable in an other lexical scope without being assigned to a reister")
				| Some r ->
				    get_display (get_var_lex v) r ^
				    "\tmovl\t" ^ (string_of_int b) ^ "(" ^ (string_of_register r) ^ "), " ^ r' ^ "\n"
			    )
			| _ -> raise (Asm_error "trying to unspill a variable not on the stack and not in the same lexical scope")
		    else
		      (
			let old_reg = v.var_reg in
			  v.var_reg <- None;
			  let r = code_of_var v in
			    v.var_reg <- old_reg;
			    "\tmovl\t" ^ r ^ ", " ^ r' ^ "\n"
		      )
		)
	  | Spill(lex_scope, r, vd) ->
	      if lex_scope then
		match !(vd.vd_pos) with
		  | Pos_base b ->
		      get_display (get_vd_lex vd) edi ^
		      "\tmovl\t" ^ (string_of_register r) ^ ", " ^ (string_of_int b) ^ "(" ^ (string_of_register edi) ^ ")\n"
		  | _ -> raise (Asm_error "trying to spill a variable not on the stack and not in the same lexical scope")
	      else
		"\tmovl\t" ^ (string_of_register r) ^ ", " ^ (string_of_stack_pos ebp (get_vd_pos vd)) ^ "\n"
	  | Call n -> "\tcall\t" ^ n ^ "\n"
	  | Set_var(f, v) ->
	      (
		match v.var_reg with
		  | Some r ->
		      if r = esi then
			let rtmp =
			  Reg_int
			    (
			      match r with
				| Reg_int rn -> if rn < 3 then rn + 1 else 0
				| _ -> raise (Asm_error "internal error")
			    )
			in
			let reg_name = string_of_char (string_of_register rtmp).[2] in
			  "\tpushl\t" ^ (string_of_register rtmp) ^ "\n" ^
			  "\tset" ^ (string_of_flag f) ^ "\t%" ^ reg_name ^ "l\n" ^
			  "\tmovzbl\t%" ^ reg_name ^ "l, " ^ (string_of_register r) ^ "\n" ^
			  "\tpopl\t" ^ (string_of_register rtmp) ^ "\n"
		      else
			let reg_name = string_of_char (string_of_register r).[2] in
			  "\tset" ^ (string_of_flag f) ^ "\t%" ^ reg_name ^ "l\n" ^
			  "\tmovzbl\t%" ^ reg_name ^ "l, " ^ (string_of_register r) ^ "\n"
		  | None -> raise (Asm_error "internal error")
	      )
	  | Nop -> "\tnop\n"
	  | Push v -> "\tpushl\t" ^ (code_of_var v) ^ "\n"
	  | Push_o(v, len) ->
	      let r = code_of_var_reg v in
		"\tsubl\t$" ^ (string_of_int len) ^ ", %esp\n" ^
		"\tmovl\t" ^ r ^ ", %esi\n" ^
		"\tmovl\t%esp, %edi\n" ^
		"\tmovl\t$" ^ (string_of_int len) ^ ", %ecx\n" ^
		"\trep movsb\n"
	  | Return v ->
	      let r = code_of_var_reg v in
		"\tmovl\t" ^ r ^ ", %eax\n" ^
		"\tmovl\t%ebp, %esp\n" ^
		"\tpopl\t%ebp\n" ^
		"\tret\n"
	  | Code c -> c
	  | Check_avail_with_flags _
	  | Check_avail _
	  | Assert_avail _
	  | Spill_var _
	  | Spill_all
	  | Read_flag _
	  | Proc _
	  | Fun _ -> raise (Asm_error "internal error in assembly generation")
    in
      
    let rec code_of_base_block bb =
      (List.fold_left (fun s i -> s ^ code_of_instr i) "" bb.bb_block) ^ code_of_link bb.bb_link;

    and code_of_link = function
      | If(v, b1, b2, b3) ->
	  let c1 = code_of_base_block b1
	  and c2 = code_of_base_block b2
	  and c3 = code_of_base_block b3 in
	  let l2 = get_fresh_label () and l3 = get_fresh_label () in
	    (
	      match v.var_def.vd_value with
		| Some(Bool b) ->
		    (if b then c1 else c2) ^ c3
		| _ ->
		    (
		      match v.var_reg with
			| Some (Reg_flag f) ->
			    "\t" ^ (jump_of_flag (neg_flag f)) ^ "\t" ^ l2 ^ "\n"
			| Some (Reg_int r) ->
			    "\tcmpl\t$0, " ^ (string_of_register (Reg_int r)) ^ "\n" ^
			    "\tje\t" ^ l2 ^ "\n"
			| None -> failwith "no value for jump's bool"
		    ) ^
		    c1 ^ "\tjmp\t" ^ l3 ^ "\n" ^
		    l2 ^ ":\n" ^ c2 ^
		    l3 ^ ":\n" ^ c3
	    )
      | While(v, be, b1, b2) ->
	  let ce = code_of_base_block be
	  and c1 = code_of_base_block b1
	  and c2 = code_of_base_block b2 in
	  let l1 = get_fresh_label () and l2 = get_fresh_label () in
	    (
	      match v.var_def.vd_value with
		| Some(Bool b) ->
		    if b then
		      l1 ^ ":\n" ^ c1 ^ "\tjmp\t" ^ l1 ^ "\n"
		    else
		      c2
		| _ ->
		    l1 ^ ":\n" ^ ce ^
		    (
		      match v.var_reg with
			| Some (Reg_flag f) ->
			    "\t" ^ (jump_of_flag (neg_flag f)) ^ "\t" ^ l2 ^ "\n"
			| Some (Reg_int r) ->
			    "\tcmpl\t$0, " ^ (string_of_register (Reg_int r)) ^ "\n" ^
			    "\tje\t" ^ l2 ^ "\n"
			| None -> failwith "no value for jump's bool"
		    ) ^
		    c1 ^ "\tjmp\t" ^ l1 ^ "\n" ^
		    l2 ^ ":\n" ^ c2
	    )
      | End -> ""
    in
      ret := !ret ^
      ".globl " ^ p.proc_name ^ "\n" ^
      "\t.type\t" ^ p.proc_name ^ ", @function\n" ^ p.proc_name ^ ":\n" ^
      "\tpushl\t%ebp\n" ^ "\tmovl\t%esp, %ebp\n" ^
      "\tmovl\t%ebp, (DISPLAY+" ^ (string_of_int (4 * p.proc_lex_depth)) ^ ")\n"; (** Let's save the ARP in the display. *)
      if p.proc_stack_size > 0 then
	ret := !ret ^ "\tsubl\t$" ^ (string_of_int p.proc_stack_size) ^ ", %esp\n";
      ret := !ret ^ code_of_base_block p.proc_value;
      if not p.proc_is_fun then
	ret := !ret ^ "\txor\t%eax, %eax\n" ^
	"\tmovl\t%ebp, %esp\n" ^
	"\tpopl\t%ebp\n" ^
	"\tret\n\n"
  in

  (** Set the positions of arguments on the stack. *)
  let set_args_pos args =
    (* 8 is because we have pushed the return adress and the caller arp on the stack *)
    ignore (List.fold_left (fun s a -> a.vd_pos := Pos_base (s + 8); s + (sizeof (get_vd_type a))) 0 args)
  in

  (** Get the size needed by the procedure on the stack. This function also sets the positions of the variables on the stack ([vd_pos]). *)
  let rec get_stack_size pv =
    List.fold_left (fun size vd ->
		      let set_on_the_stack () =
			vd.vd_pos := Pos_base (-size - (sizeof (get_vd_type vd)));
			size + (sizeof (get_vd_type vd))
		      in
			match (get_vd_pos vd), get_vd_type vd with
			  | Pos_none, _ when (get_vd_ots vd) (* && not (is_constant (new_var v)) *) ->
			      set_on_the_stack ()
			  | _, Other _ ->
			      set_on_the_stack ()
			  | _ -> size
		   ) 0 pv
  in
    List.iter
      (fun p ->
	 set_args_pos p.proc_args;
	 allocate_registers p.proc_value p.proc_lex_depth;
	 if !Params.local_opts then
	   local_opt_base_block p.proc_value;
	 p.proc_stack_size <- get_stack_size p.proc_vars
      ) prog.prog_procs;
    ret := "\t.file\t\"" ^ prog.prog_source_name ^ "\"\n" ^
    "\t.section\t.data\nDISPLAY:\n\t.zero\t" ^ (string_of_int (4 * (prog.prog_max_depth + 1))) ^ "\n";
    ret := !ret ^ "\t.text\n";
    List.iter (fun p -> code_of_proc p) prog.prog_procs; !ret ^ "\t.size\tmain, .-main\n" ^ "\t.section\t.note.GNU-stack,\"\",@progbits\n" ^ "\t.ident\t\"Bifton 0.1.0\"\n"
