%{
  (*
    Bifton - A Pascal- compiler.

    Copyright (C) 2004 Samuel Mimram & Julien Cristau.
  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *)

  (**
    Pascal- parser.

    @author Samuel Mimram
  *)

  (* $Id: parser.mly,v 1.20 2004/04/29 11:18:57 smimou Exp $ *)

  (*
    TODO:
    - comments
    - if ... then [no else]
    - put the Compound back to its place (??)
  *)

  open Pascal_ir

  (**
    Get current position (line, column).
  *)
  let get_cur_pos () =
    let pos = Parsing.symbol_start_pos () in
      pos.Lexing.pos_lnum, (pos.Lexing.pos_cnum - pos.Lexing.pos_bol)

  (**
    A statement.
  *)
  let new_statement stt =
    {
      stt_value = stt;
      stt_pos = get_cur_pos ();
      stt_asm = None;
    }

  (**
    A variable (or procedure, function, etc) use.
  *)
  let new_variable name =
    {
      var_name = name;
      var_type = Unknown_type;
      var_def = None;
      var_nature = Unknown_nature;
      var_pos = get_cur_pos ();
    }

  (**
    A variable dclaration (specifying its initial value).
  *)
  let new_variable_val_def value name t byref =
    {
      def_name = name;
      def_type = t;
      def_nature = Varb;
      def_value = value;
      def_pos = get_cur_pos ();
      def_coord = [];
      def_asm = Asm_none;
      def_pass_by_ref = byref;
    }

  (**
    A variable declaration.
  *)
  let new_variable_def = new_variable_val_def Variable

  (**
    A type declaration.
  *)
  let new_type_def name t =
    {
      def_name = name;
      def_type = Type_type;
      def_nature = Cst;
      def_value = Type t;
      def_pos = get_cur_pos ();
      def_coord = [];
      def_asm = Asm_none;
      def_pass_by_ref = false;
    }

  (**
    A constant declaration.
  *)
  let new_const_def name v =
    {
      def_name = name;
      def_type = Unknown_type;
      def_nature = Cst;
      def_value = Constant v;
      def_pos = get_cur_pos ();
      def_coord = [];
      def_asm = Asm_none;
      def_pass_by_ref = false;
    }

  (**
    An expression.
  *)
  let new_expression value =
    {
      expr_val = value;
      expr_type = Unknown_type;
      expr_pos = get_cur_pos ();
    }

  let expr_of_var_def vd =
    new_expression (Var(new_variable vd.def_name, []))

  (**
    We are handling errors.
  *)
  let parse_error err =
    raise (Parse_error(err, get_cur_pos ()))

  (**
    Create a new block.
  *)
  let new_block defs statements =
    {
      block_env = concat_env true (env_of_def_list defs) (create_environment ());
      block_stts = statements;
    }
%}

%token PROGRAM CONST TYPE ARRAY OF RECORD END VAR IF THEN ELSE WHILE DO BEGIN END RETURN PROCEDURE FUNCTION NEW DELETE
%token TRUE FALSE NOT
%token INTEGER BOOLEAN POINTER
%token LBRACK RBRACK INTERV SEMICOL COMMA COLON DOT LPAR RPAR ASSIGN ADDRESS_OF
%token EQ NEQ LE LT GE GT
%token PLUS MINUS OR MULT DIV MOD AND
%token <int> NUMERAL
%token <string> NAME

%nonassoc RETURN
%left AND OR
%nonassoc LT LE GT GE EQ NEQ
%nonassoc NOT
%left MOD
%left PLUS MINUS
%left MULT DIV
%nonassoc UPLUS UMINUS
%nonassoc NEW DELETE
%nonassoc POINTER
%nonassoc ADDRESS_OF
%nonassoc LBRACK RBRACK
%left DOT

%nonassoc NOELSE
%nonassoc ELSE

%start program
%type <Pascal_ir.program> program
%%

  name_comma_list:
| NAME COMMA name_comma_list { $1 :: $3 }
| NAME { [$1] }
  ;

  variable: NAME { new_variable $1 }
  ;

  program: PROGRAM NAME SEMICOL block_body DOT { ($2, $4) }
  ;

  block_body: constant_definition_part type_definition_part variable_definition_part procedure_definitions compound_statement { new_block ($1 @ $2 @ $3 @ $4) $5 }
  ;

  constant_definition_part:
| CONST constant_definitions { $2 }
| { [] }
  ;

  constant_definitions:
| NAME EQ expression SEMICOL constant_definitions { new_const_def $1 $3 :: $5 }
| { [] }
  ;

  type_definition_part:
| TYPE type_definitions { $2 }
| { [] }
  ;

  type_definitions:
| NAME EQ new_type SEMICOL type_definitions { (new_type_def $1 $3) :: $5 }
| { [] }
  ;

  new_type:
| new_array_type { $1 }
| new_record_type { $1 }
  ;

  new_array_type: ARRAY LBRACK constant_name_use INTERV constant_name_use RBRACK OF type_name_use { Array ($8, $3, $5) }
  ;

  constant_name_use: expression { $1 }
  ;

  type_name_use:
| INTEGER { Integer }
| BOOLEAN { Boolean }
| POINTER type_name_use { Pointer $2 }
| NAME { User_type $1 }
  ;

  new_record_type: RECORD record_section field_list END { Record ($2 @ $3) }
  ;

  field_list:
| SEMICOL record_section field_list { $2 @ $3 }
| { [] }
  ;

  record_section: name_comma_list COLON type_name_use { List.map (fun n -> (n, $3)) $1 }
  ;

  variable_definition_part:
| VAR variable_definitions { $2 }
| { [] }
  ;

  variable_definitions:
| variable_definition variable_definitions { $1 @ $2 }
| { [] }
  ;

  variable_definition: name_comma_list COLON type_name_use SEMICOL { List.map (fun n -> new_variable_def n $3 false) $1 }
  ;

  expression: expr { new_expression $1 }
  ;

  expr:
| expression LT expression { Lt($1, $3) }
| expression LE expression { Le($1, $3) }
| expression EQ expression { Eq($1, $3) }
| expression NEQ expression { Neq($1, $3) }
| expression GE expression { Le($3, $1) }
| expression GT expression { Lt($3, $1) }
| PLUS expression %prec UPLUS { $2.expr_val }
| MINUS expression %prec UMINUS { Uminus $2 }
| expression PLUS expression { Plus($1, $3) }
| expression MINUS expression { Minus($1, $3) }
| expression OR expression { Or($1, $3) }
| expression MULT expression { Mult($1, $3) }
| expression DIV expression { Div($1, $3) }
| expression MOD expression { Mod($1, $3) }
| expression AND expression { And($1, $3) }
| NUMERAL { Int $1 }
| TRUE { Bool true }
| FALSE { Bool false }
| variable_access { $1 }
| variable LPAR actual_parameters RPAR { Var($1, $3) }
| LPAR expression RPAR { $2.expr_val }
| NOT expression { Not $2 }
| NEW type_name_use { New(ref $2) }
  ;

  // name use: constant / variable
    variable_access:
| variable { Var ($1, []) }
| variable_access POINTER { Bang (new_expression $1) }
//| ADDRESS_OF variable_access { Address_of(new_expression $2) }  // TODO: why is that shift reducing ????
| expression LBRACK expression RBRACK { Arr($1, $3) }
| expression DOT NAME { Rec($1, $3) }
  ;

  statement: stt { new_statement $1 };

  stt:
| assignment_statement { $1 }
| procedure_statement { $1 }
| if_statement { $1 }
| while_statement { $1 }
| compound_statement { Compound $1 }
| delete_statement { $1 }
| return_statement { $1 }
| { Noop }
  ;


  return_statement:
| RETURN expression { Return $2 }
  ;

  delete_statement:
| DELETE variable_access { Delete (new_expression $2) }
  ;

  assignment_statement:
| variable_access ASSIGN expression { Assign(new_expression $1, $3) }
  ;

  procedure_statement:
| variable { Proc($1, []) }
| variable LPAR actual_parameters RPAR { Proc($1, $3) }
  ;

  actual_parameters:
| actual_parameter COMMA actual_parameters { $1 :: $3 }
| actual_parameter { [$1] }
  ;

  actual_parameter: expression { $1 };

  if_statement:
| IF expression THEN statement %prec NOELSE { If($2, $4, new_statement Noop) }
| IF expression THEN statement ELSE statement { If($2, $4, $6) }
  ;

  while_statement: WHILE expression DO statement { While($2, $4) };

  compound_statement: BEGIN statement statements END { $2 :: $3 };

  statements:
| SEMICOL statement statements { $2 :: $3 }
| { [] }
  ;

  procedure_definitions:
| PROCEDURE NAME procedure_block SEMICOL procedure_definitions
      {
	let args, block = $3 in
	  (new_function $2 args Unit block (get_cur_pos ()) false) :: $5
      }
| FUNCTION NAME function_block SEMICOL procedure_definitions
      {
	let args, rett, block = $3 in
	  (new_function $2 args rett block (get_cur_pos ()) false) :: $5
      }
| { [] }
  ;

  procedure_block:
| SEMICOL block_body { ([], $2) }
| LPAR parameter_definitions RPAR SEMICOL block_body
      {
	let block = $5 in
	  List.iter
	    (function
	       | _, None -> ()
	       | v, Some br ->
		   block.block_env <- add_def block.block_env br.def_name br;
		   block.block_stts <-
		   new_statement (Assign(expr_of_var_def br,
					 new_expression (Bang(expr_of_var_def v)))) ::
		     block.block_stts @
		   [new_statement (Assign(new_expression (Bang(expr_of_var_def v)), expr_of_var_def br))]
	    ) $2;
	  (List.map fst $2, block)
      }
  ;

  function_block:
| COLON type_name_use SEMICOL block_body { ([], $2, $4) }
| LPAR parameter_definitions RPAR COLON type_name_use SEMICOL block_body
      {
	let block = $7 in
	  List.iter
	    (function
	       | _, None -> ()
	       | v, Some br ->
		   block.block_env <- add_def block.block_env br.def_name br;
		   block.block_stts <-
		   new_statement (Assign(expr_of_var_def br,
					 new_expression (Bang(expr_of_var_def v)))) ::
		     block.block_stts @
		   [new_statement (Assign(new_expression (Bang(expr_of_var_def v)), expr_of_var_def br))]
	    ) $2;
	  (List.map fst $2, $5, block)
      }
  ;

  parameter_definitions:
| parameter_definition SEMICOL parameter_definitions { $1 @ $3 }
| parameter_definition { $1 }
  ;

// The second member of the elements of the list is for variables passed by ref (we must add the code to dereference its pointer and store it after use)
  parameter_definition:
| VAR name_comma_list COLON type_name_use
      {
	List.map
	 (fun v ->
	    new_variable_def ("@" ^ v) (Pointer $4) true,
	    Some (new_variable_def v $4 false)
         ) $2
      }
| name_comma_list COLON type_name_use { List.map (fun v -> new_variable_def v $3 false, None) $1 }
  ;
