(*
  Bifton - A Pascal- compiler.

  Copyright (C) 2004 Samuel Mimram & Julien Cristau.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(**
  Abstract syntax tree with a Pascal looking-like structure.

  @author Samuel Mimram
*)

(* $Id: pascal_ir.ml,v 1.29 2004/05/09 19:32:58 smimou Exp $ *)

(*
  TODO:
  - mli
  - replace errors with generic functions
  - s/proc/func/
*)

(**
  A parsing error occured. The arguments are the column and line numbers.
*)
exception Parse_error of string * (int * int)

(**
  Types of values.
*)
type pascal_type =
  | Integer (** an integer *)
  | Boolean (** a boolean *)
  | Record of (string * pascal_type) list (** a record *)
  | Array of (pascal_type * expression * expression) (** an array *)
  | User_type of string (** a user type *)
  | Unit (** unit (not used for now) *)
  | Pointer of pascal_type (** a pointer *)
  | Type_type (** a type *)
  | Unknown_type (** no type yet *)

(**
  Nature.
*)
and nature =
  | Varb (** a variable *)
  | Cst (** a constant *)
  | Funct (** a function *)
  | Unknown_nature (** no nature yet *)

(**
  A reference to an identifier. [var_by_name] is only meaningful for procedures' and functions' arguments (yet) and should be set to [false] else.
*)
and variable =
    {
      var_name : string; (** name of the variable *)
      mutable var_type : pascal_type; (** type of the variable *)
      mutable var_def : definition option; (** definition *)
      mutable var_nature : nature; (** nature *)
      var_pos : int * int; (** position in the source code *)
    }

(**
  A typed expression.
*)
and expression = 
    {
      expr_val : expr; (** the actual expression *)
      mutable expr_type : pascal_type; (** type of the expression *)
      expr_pos: int * int; (** position in the source code (line, column) *)
    }

(**
  An expression.
*)
and expr =
  | Lt of expression * expression (** less than *)
  | Le of expression * expression (** less or equal *)
  | Eq of expression * expression (** equal *)
  | Neq of expression * expression (** not equal *)
  | Uminus of expression (** unary minus *)
  | Plus of expression * expression (** plus *)
  | Minus of expression * expression (** minus *)
  | Or of expression * expression (** logical or *)
  | Mult of expression * expression (** times *)
  | Div of expression * expression (** divided by *)
  | Mod of expression * expression (** modulo *)
  | And of expression * expression (** logical and *)
  | Not of expression (** logical not *)
  | Int of int (** an integer *)
  | Bool of bool (** a boolean *)
  | Var of variable * expression list (** a variable or a function call *)
  | Arr of expression * expression (** an array *)
  | Rec of expression * string (** a member of a structure *)
  | Address_of of expression (** the address of a variable *)
  | Bang of expression (** the value pointed by a variable *)
  | New of pascal_type ref (** a new value of this type (malloc) *)

(**
  A statement.
*)
and stt =
  | Assign of expression * expression (** assignation of a value to a variable *)
  | Proc of variable * expression list (** procedure call *)
  | If of expression * statement * statement (** if ... then ... else ... (if there is no else then the second [statement] should be [Noop]) *)
  | While of expression * statement (** while *)
  | Compound of statement list (** begin ... end *)
  | Return of expression (** return *)
  | Delete of expression (** delete a value (free) *)
  | Noop (** take some rest *)

(**
  A decorated statement.
*)
and statement =
    {
      stt_value : stt; (** the actual statement *)
      stt_pos : int * int; (** position in source code *)
      mutable stt_asm : (Asm_ir.variable_def list * Asm_ir.meta_instr) option; (** for 'if' and 'while' *)
    }

(**
  An environment (a list of names and their value).
*)
and environment = (string * definition) list

(**
  A block (i.e. a list of statements in an environment).
*)
and block =
    {
      mutable block_env : environment; (** the preliminary variables declarations *)
      mutable block_stts : statement list; (** the instructions *)
    }

(**
  A procedure or a function.
*)
and func =
    {
      func_args : definition list; (** arguments *)
      func_value : block; (** code *)
      func_asm_def : Asm_ir.procedure; (** definition in the asm ir *)
      mutable func_used : bool; (** for dead-code elimination purposes *)
      mutable func_external : bool; (** is the procedure external (defined in bifton's standard library) *)
    }

(**
  What an identifier can be.
*)
and value =
  | Constant of expression (** a constant expression *)
  | Variable (** a varaiable containing a value (not passed by reference) *)
  | Function of func (** a function *)
  | Type of pascal_type (** a type *)

(**
  Definition in the assembly side.
*)
and asm_value =
  | Asm_var of Asm_ir.variable_def (** a variable *)
  | Asm_proc of Asm_ir.procedure (** a procedure *)
  | Asm_none (** no useful assembly is associated to this value *)

(**
  An identifier definition.
*)
and definition =
    {
      def_name : string; (** name *)
      mutable def_type : pascal_type; (** type *)
      def_nature : nature; (** nature *)
      mutable def_value : value; (** value *)
      def_pos : int * int; (** position in the source code *)
      mutable def_coord : int list; (** lexical position *)
      mutable def_asm : asm_value; (** translation into asm *)
      def_pass_by_ref : bool; (** should the value be passed by reference (for arguments of functions)? *)
    }

(** {2 Operations on environments.} *)

(**
  A variable is being redefined.
*)
exception Already_defined

(**
  Create a new empty environment.
*)
let create_environment () : environment = []

(**
  Fold a function throughout an environment.
*)
let fold_env f x (env:environment) =
  List.fold_left (fun x (n, v) -> f n v x) x env

(**
  Iter a function on all elements of an environment.
*)
let iter_env f (env:environment) =
  List.iter (fun (n, v) -> f n v) env

(**
  Get the environment corresponding to a list of variable definitions.
*)
let env_of_def_list =
    List.map (fun d -> (d.def_name, d))

(**
  Add a definition to an environment.

  @raise Already_defined if a previous definition was already made with the same name.
*)
let rec add_def (env:environment) name def =
  try
    ignore (get_def env name);
    raise Already_defined
  with
    | Not_found -> (name, def) :: env

(**
  Get a the value associated to a name in an environment. Remark: don't forget that procedures and functions are binders and should therefore be added to the current environment.

  @raise Not_found if no definition was made for that name.
*)
and get_def (env:environment) name =
  List.assoc name env

(**
  Concatenate two environments. The first argument determines whereas the function should check for duplicate definitions or not.
*)
let concat_env check_dup (env1:environment) env2 =
  if not check_dup then
    env1 @ env2
  else
    fold_env
      (fun n d env ->
	 try
	   add_def env n d
	 with
	   | Already_defined -> raise (Parse_error ("variable " ^ d.def_name ^ " was defined twice", d.def_pos))
      ) env2 (List.rev env1)


(** {2 Operations on programs.} *)


(**
  Create a new function.
*)
let new_function name args t value pos ext =
  {
    def_name = name;
    def_type = t;
    def_nature = Funct;
    def_value =
     Function (
       let p = 
	 {
	   func_args = args;
	   func_value = value;
	   func_used = false;
	   func_asm_def = Asm_ir.proc_stub name (List.map (fun a -> Asm_ir.new_var_def Asm_ir.No_type (-1)) args) 0; (* TODO: globally unique name *)
	   func_external = ext;
	 }
       in
	 List.iter2 (fun a b -> a.def_asm <- Asm_var b) args p.func_asm_def.Asm_ir.proc_args;
	 (** An argument is considered as any other variable. *)
	 value.block_env <- concat_env true (env_of_def_list args) value.block_env;
	 p
     );
    def_pos = pos;
    def_coord = [];
    def_asm = Asm_none;
    def_pass_by_ref = false;
  }

(**
  Get a human readable representation of a type.
*)
let rec string_of_type = function
  | Integer -> "integer"
  | Boolean -> "boolean"
  | Pointer t -> "^" ^ (string_of_type t)
  | User_type t -> t
  | Unit -> "unit"
  | Array(t, _, _) -> "array [" ^ (string_of_type t) ^ "]"
  | Record l -> (List.fold_left (fun s (n, t) -> s ^ " " ^ n ^ " (" ^ (string_of_type t) ^ ")") "record [" l) ^ "]"
  | Type_type -> "type"
  | Unknown_type -> "???"

(**
  A program.
*)
type program = string * block

(**
  Is a value equal to Some _?
*)
let is_some = function
  | Some _ -> true
  | None -> false

(**
  Get the value [v] in a value equal to [Some v].
*)
let get_some = function
  | Some s -> s
  | None -> raise Not_found

(**
  Compute the value of a constant integer expression.
*)
let rec eval_expr_int e =
  match e.expr_val with
    | Int i -> i
    | Var(v, []) when is_some v.var_def ->
	(
	  match (get_some v.var_def).def_value with
	    | Constant c -> eval_expr_int c
	    | _ -> raise (Parse_error("internal error in eval_expr_int function (trying to compute a non-integer expression)", e.expr_pos))
	)
    | Uminus e -> -(eval_expr_int e)
    | Plus(e1, e2) -> (eval_expr_int e1) + (eval_expr_int e2)
    | Minus(e1, e2) -> (eval_expr_int e1) - (eval_expr_int e2)
    | Mult(e1, e2) -> (eval_expr_int e1) * (eval_expr_int e2)
    | Div(e1, e2) -> (eval_expr_int e1) / (eval_expr_int e2)
    | Mod(e1, e2) -> (eval_expr_int e1) mod (eval_expr_int e2)
    | Not _ | And _ | Or _ | Neq _ | Eq _ | Le _ | Lt _ | Bool _
	-> raise (Parse_error("internal error in eval_expr_int function (trying to compute a non-integer expression)", e.expr_pos))
    | Bang _ | Address_of _ | Rec _ | Arr _ | Var _ | New _
	-> raise (Parse_error("internal error in eval_expr_int function (trying to compute a non-constant expression)", e.expr_pos))

(**
  Compute the value of a constant boolean expression.
*)
and eval_expr_bool e =
  match e.expr_val with
    | Bool b -> b
    | Var(v, []) when is_some v.var_def ->
	(
	  match (get_some v.var_def).def_value with
	    | Constant c -> eval_expr_bool c
	    | _ -> raise (Parse_error("internal error in eval_expr_bool function (trying to compute a non-boolean expression)", e.expr_pos))
	)
    | Le(e1, e2) -> (eval_expr_int e1) <= (eval_expr_int e2)
    | Lt(e1, e2) -> (eval_expr_int e1) < (eval_expr_int e2)
    | Eq(e1, e2) ->
	(
	  match e1.expr_type with
	    | Integer -> (eval_expr_int e1) = (eval_expr_int e2)
	    | Boolean -> (eval_expr_bool e1) = (eval_expr_bool e2)
	    | _ -> raise (Parse_error("internal error in eval_expr_bool function (equality can only be computed between integers or booleans)", e.expr_pos))
	)
    | Neq(e1, e2) ->
	(
	  match e1.expr_type with
	    | Integer -> (eval_expr_int e1) <> (eval_expr_int e2)
	    | Boolean -> (eval_expr_bool e1) <> (eval_expr_bool e2)
	    | _ -> raise (Parse_error("internal error in eval_expr_bool function (inequality can only be computed between integers or booleans)", e.expr_pos))
	)
    | Not e -> not (eval_expr_bool e)
    | Or(e1, e2) -> (eval_expr_bool e1) || (eval_expr_bool e2)
    | And(e1, e2) -> (eval_expr_bool e1) && (eval_expr_bool e2)
    | _ -> raise (Parse_error("internal error in eval_expr_bool function (trying to compute a non-boolean expression)", e.expr_pos))

(**
  Compute the value of a constant expression.
*)
let eval_expr e =
  match e.expr_type with
    | Integer -> Asm_ir.Int (eval_expr_int e)
    | Boolean -> Asm_ir.Bool (eval_expr_bool e)
    | _ -> raise (Parse_error("internal error in eval_expr function (trying to compute a non-boolean-or-integer expression)", e.expr_pos))

(**
  An error occured during the pascal-to-asm conversion.
*)
exception Asm_error of string

(**
  Get the size needed by a type (in bytes).
*)
let rec sizeof = function
  | Pointer _
  | Integer -> 4
  | Boolean -> 1
  | Record r -> List.fold_left (fun s (_, t) -> s + (sizeof t)) 0 r
  | Array(t, d, f) -> (sizeof t) * ((eval_expr_int f) - (eval_expr_int d) + 1)
  | t ->
      Log.debug (string_of_type t);
      raise (Asm_error "internal error in sizeof function")

(**
  Get the asm type corresponding to a pascal type.
*)
let asm_type = function
  | Integer -> Asm_ir.Integer
  | Boolean -> Asm_ir.Boolean
  | Pointer _ -> Asm_ir.Pointer
  | Array _ as a -> Asm_ir.Other(sizeof a)
  | Record r -> Asm_ir.Other(List.fold_left (fun s (_, t) -> s + (sizeof t)) 0 r)
  | User_type _
  | Unknown_type
  | Type_type
  | Unit -> raise (Parse_error("internal error in asm_type function (typing inconsistency)", (-1, -1)))

(**
  Do several checks on a block:
  - check that called variables
  - check data types
  @param rettype the type of the returned value
  @raise Parse_error if an error was detected.
*)
let rec check_block block block_ctx rettype =
  let typing_error has_t used_with_t pos =
    raise (Parse_error("typing error (expression has type " ^ (string_of_type has_t) ^ " but is used with type " ^ (string_of_type used_with_t) ^ ")", pos))
  in
  let undefined_variable_error v =
    raise (Parse_error("variable " ^ v.var_name ^ " is used without having been declared", v.var_pos))
  in
    (** Find a definition in a context (i.e. an environment list). *)
  let rec get_def_ctx ctx name =
    match ctx with
      | c :: s ->
	  (
	    try
	      get_def c name
	    with
	      | Not_found -> get_def_ctx s name
	  )
      | [] -> raise Not_found
  in
    (** Get the definition of a variable. It uses / fills in var_def. *)
  let get_var_def ctx v =
    match v.var_def with
      | Some d -> d
      | None ->
	  try
	    let d = get_def_ctx ctx v.var_name in
	      v.var_def <- Some d; d
	  with
	    | Not_found -> undefined_variable_error v
  in
  let get_def_value ctx name =
    (get_def_ctx ctx name).def_value
  in
  let get_var_value ctx v =
    (get_var_def ctx v).def_value
  in
    (** Get the type of a variable. It infers it if it was not done before. *)
  let rec get_var_type ctx v =
    if v.var_type = Unknown_type then
      v.var_type <- (get_var_def ctx v).def_type;
    v.var_type
      (** Get the nature of a variable. It infers it if it was not done before. *)
  and get_var_nature ctx v =
    if v.var_nature = Unknown_nature then
      v.var_nature <- (get_var_def ctx v).def_nature;
    v.var_nature
  and is_constant = function
    | Cst -> true
    | _ -> false
	(** Get the nature of an expression. It infers it if it was not done before. *)
  and get_expr_nature ctx e =
    let is_a_constant v =
      match (get_var_def ctx v).def_value with
	| Constant _ -> true
	| _ -> false
    in
    let score_of_nature = function
      | Cst -> 0
      | Varb -> 1
      | Funct -> 2
      | Unknown_nature -> failwith "trying to get the nature of a not yet typed expression"
    in
    let nature_of_score = function
      | 0 -> Cst
      | 1 -> Varb
      | 2 -> Funct
      | _ -> failwith "internal error"
    in
      match e.expr_val with
	| Lt(e1, e2)
	| Le(e1, e2)
	| Eq(e1, e2)
	| Neq(e1, e2)
	| Plus(e1, e2)
	| Minus(e1, e2)
	| Mult(e1, e2)
	| Div(e1, e2)
	| Mod(e1, e2)
	| And(e1, e2)
	| Or(e1, e2) ->
	    nature_of_score (max (score_of_nature (get_expr_nature ctx e1)) (score_of_nature (get_expr_nature ctx e2)))
	| Uminus e
	| Not e -> get_expr_nature ctx e
	| Int _
	| Bool _ -> Cst
	| Var(v, []) when is_a_constant v -> Cst
	| Var(v, _) -> Varb
	| New _
	| Bang _
	| Address_of _
	| Rec _
	| Arr _ -> Varb (* first approximation *)

  and resolve_user_type ctx t =
    let get_user_type_def tn =
      try
	match get_def_value ctx tn with
	  | Type t -> t
	  | _ -> raise (Parse_error(tn ^ " is not an user type", (-1, -1)))
      with
	| Not_found ->
	    raise (Parse_error("user type " ^ tn ^ " is used without having been declared", (-1, -1)))
    in
      match t with
	| User_type tn -> get_user_type_def tn
	| Array(t, i, f) -> Array(resolve_user_type ctx t, i, f)
	| Record l -> Record((List.map (fun (n, t) -> n, resolve_user_type ctx t)) l)
	| Pointer t -> Pointer (resolve_user_type ctx t)
	| _ -> t

  (** Get the type of an expression. It infers it if it was not done before. *)
  and get_expr_type ctx e =
    if e.expr_type = Unknown_type then
      e.expr_type <- resolve_user_type ctx
	(
	  match e.expr_val with
	    | Lt(e1, e2)
	    | Le(e1, e2) ->
		let te1 = get_expr_type ctx e1 in
		let te2 = get_expr_type ctx e2 in
		  if te1 <> Integer then
		    typing_error te1 Integer e1.expr_pos
		  else if te2 <> Integer then
		    typing_error te2 Integer e2.expr_pos
		  else Boolean
	    | Eq(e1, e2)
	    | Neq(e1, e2) ->
		let te1 = get_expr_type ctx e1 in
		let te2 = get_expr_type ctx e2 in
		  if te1 <> Integer && te1 <> Boolean then
		    typing_error te1 te2 e1.expr_pos;
		  if te1 <> te2 then
		    typing_error te1 te2 e1.expr_pos;
		  Boolean
	    | Plus(e1, e2)
	    | Minus(e1, e2)
	    | Mult(e1, e2)
	    | Div(e1, e2)
	    | Mod(e1, e2) ->
		let te1 = get_expr_type ctx e1 in
		let te2 = get_expr_type ctx e2 in
		  if te1 <> Integer then
		    typing_error te1 Integer e1.expr_pos
		  else if te2 <> Integer then
		    typing_error te2 Integer e2.expr_pos
		  else Integer
	    | Uminus e ->
		let te = get_expr_type ctx e in
		  if te <> Integer then
		    typing_error te Integer e.expr_pos
		  else Integer
	    | And(e1, e2)
	    | Or(e1, e2) ->
		let te1 = get_expr_type ctx e1 in
		let te2 = get_expr_type ctx e2 in
		  if te1 <> Boolean then
		    typing_error te1 Boolean e1.expr_pos
		  else if te2 <> Boolean then
		    typing_error te2 Boolean e2.expr_pos
		  else Boolean
	    | Not e ->
		let te = get_expr_type ctx e in
		  if te <> Boolean then
		    typing_error te Boolean e.expr_pos
		  else Boolean
	    | Int _ -> Integer
	    | Bool _ -> Boolean
	    | Arr(e1, e2) ->
		if get_expr_type ctx e2 <> Integer then
		  raise (Parse_error("typing error (integer expected)", e2.expr_pos))
		else
		  (
		    match get_expr_type ctx e1 with
		      | Array(t, _, _) -> t
		      | t -> raise (Parse_error("typing error (array expected)", e.expr_pos))
		  )
	    | Rec(e1, m) ->
		(
		  match get_expr_type ctx e1 with
		    | Record l ->
			(
			  try
			    List.assoc m l
			  with
			    | Not_found -> raise (Parse_error("no such member", e1.expr_pos))
			)
		    | _ -> raise (Parse_error("typing error (record expected)", e1.expr_pos))
		)
	    | Var(v, []) ->
		get_var_type ctx v
	    | Var(v, a) ->
		(
		  try
		    match get_var_value ctx v with
		      | Function f ->
			  (
			    try
			      List.iter2
				(fun d e ->
				   if
				     (
				       if d.def_pass_by_ref then
					 match d.def_type with
					   | Pointer t -> t
					   | _ -> raise (Parse_error("internal error (pointer expected for by ref)", v.var_pos))
				       else
					 d.def_type
				     )
				     <> get_expr_type ctx e then
				       raise (Parse_error("typing error", e.expr_pos))
				) f.func_args a;
			      get_var_type ctx v
			    with
			      | Invalid_argument _ -> raise (Parse_error("wrong number of arguments", v.var_pos))
			  );
		      | _ -> raise (Parse_error("typing error (function expected)", v.var_pos))
		  with
		    | Not_found ->
			raise (Parse_error("procedure " ^ v.var_name ^ " is used without having been declared", v.var_pos))
		)
	    | Address_of e ->
		Pointer (get_expr_type ctx e)
	    | Bang e ->
		(
		  match get_expr_type ctx e with
		    | Pointer t -> t
		    | _ -> raise (Parse_error("typing error (pointer expected)", e.expr_pos))
		)
	    | New t ->
		t := resolve_user_type ctx !t;
		Pointer !t
	);
    e.expr_type
  in
    (** Check a statement list in a context. *)
  let rec check_statement ctx stt =
    match stt.stt_value with
      | Assign(n, e) ->
	  let tn = get_expr_type ctx n
	  and te = get_expr_type ctx e in
	    if tn <> te then typing_error te tn e.expr_pos
      | Compound l -> List.iter (check_statement ctx) l
      | While(e, s) ->
	  if get_expr_type ctx e <> Boolean then
	    raise (Parse_error("typing error (boolean expected)", e.expr_pos))
	  else
	    check_statement ctx s
      | If(e, s1, s2) ->
	  if get_expr_type ctx e <> Boolean then
	    raise (Parse_error("typing error (boolean expected)", e.expr_pos))
	  else
	    (
	      check_statement ctx s1;
	      check_statement ctx s2
	    )
      | Return(e) ->
	  if get_expr_type ctx e <> rettype then
	    raise (Parse_error("typing error (" ^ string_of_type rettype ^ " expected)", e.expr_pos))
      | Proc(v, l) ->
	  (
	    try
	      match get_var_value ctx v with
		| Function f ->
		    (
		      try
			List.iter2
			  (fun d e ->
			     if
			       (
				 if d.def_pass_by_ref then
				   match d.def_type with
				     | Pointer t -> t
				     | _ -> raise (Parse_error("internal error (pointer expected for by ref)", v.var_pos))
				 else
				   d.def_type
			       )
			       <> get_expr_type ctx e then
				 raise (Parse_error("typing error", e.expr_pos))
			  ) f.func_args l;
			if get_var_type ctx v <> Unit then
			  Log.warning ("Discading the value of a function line " ^ (string_of_int (fst stt.stt_pos)))
		      with
			| Invalid_argument _ -> raise (Parse_error("wrong number of arguments", stt.stt_pos))
		    );
		| _ -> raise (Parse_error("typing error (procedure expected)", stt.stt_pos))
	    with
	      | Not_found ->
		  raise (Parse_error("procedure " ^ v.var_name ^ " is used without having been declared", v.var_pos))
	  )
      | Delete _ (* TODO: we could check that it is allocated by a new *)
      | Noop -> ()
  in
  let ctx = ref block_ctx in
  let cur_env = ref (create_environment ()) in
  let resolve_user_type t = resolve_user_type !ctx t in    
  let resolve_def_user_type d =
    d.def_type <-
    (
      try
	resolve_user_type d.def_type
      with
	| Parse_error(e, _) -> raise (Parse_error(e, d.def_pos))
    )
  in
    iter_env
      (fun n d ->
	 (
	   match d.def_value with
	     | Type t ->
		 (
		   try
		     d.def_value <- Type(resolve_user_type t)
		   with
		     | Parse_error(e, _) -> raise (Parse_error(e, d.def_pos))
		 )
	     | _ -> ()
	 );
	 resolve_def_user_type d;
	 (
	   match d.def_value with
	     | Function f ->
		 List.iter resolve_def_user_type f.func_args;
		 List.iter2 (fun a b -> b.Asm_ir.vd_type := asm_type a.def_type) f.func_args f.func_asm_def.Asm_ir.proc_args;
		 check_block f.func_value (add_def !cur_env n d :: !ctx) d.def_type (* TODO: is that the good environment? *)
	     | Constant e ->
		 if not (is_constant (get_expr_nature !ctx e)) then
		   raise (Parse_error("typing error (constant expression expected)", e.expr_pos));
		 d.def_type <- get_expr_type !ctx e
	     | Type t ->
		 (
		   match t with
		     | Array(t, i, f) ->
			 if get_expr_type !ctx i <> Integer then
			   raise (Parse_error("typing error (integer expected)", i.expr_pos))
			 else if get_expr_type !ctx f <> Integer then
			   raise (Parse_error("typing error (integer expected)", f.expr_pos))
			 else if not (is_constant (get_expr_nature !ctx i)) then
			   raise (Parse_error("typing error (constant expression expected)", i.expr_pos))
			 else if not (is_constant (get_expr_nature !ctx f)) then
			   raise (Parse_error("typing error (constant expression expected)", f.expr_pos))
		     | Record _ -> ()
		     | _ -> raise (Parse_error("internal error (user type other than array or record defined)", d.def_pos))
		 )
	     | Variable -> ()
	 );
	 cur_env := add_def !cur_env n d;
	 ctx := !cur_env :: block_ctx
      ) block.block_env;
    List.iter (check_statement (block.block_env :: block_ctx)) block.block_stts

(**
  Mark [def_used] of used procedures.
*)
let opt_useless program =
  let rec check_expr e =
    match e.expr_val with
      | Var(f, l) ->
	  (
	    match f.var_def with
	      | None -> failwith "internal error"
	      | Some d ->
		  List.iter check_expr l;
		  match d.def_value with
		    | Function f ->
			if not f.func_used then
			  (
			    f.func_used <- true;
			    check_block f.func_value
			  )
		    | _ -> () (** This is a variable. *)
	  )
      | Arr(e1, e2)
      | Lt(e1, e2)
      | Le(e1, e2)
      | Eq(e1, e2)
      | Neq(e1, e2)
      | Or(e1, e2)
      | And(e1, e2)
      | Plus(e1, e2)
      | Minus(e1, e2)
      | Mult(e1, e2)
      | Div(e1, e2)
      | Mod(e1, e2) ->
	  check_expr e1;
	  check_expr e2
      | Rec(e, _)
      | Address_of e
      | Bang e
      | Not e
      | Uminus e ->
	  check_expr e
      | New _
      | Int _
      | Bool _ -> ()
  and check_instr = function
    | Proc(p, l) ->
	let d =
	  match p.var_def with
	    | Some d -> d
	    | None -> failwith "internal error"
	in
	  (
	    List.iter check_expr l;
	    match d.def_value with
	      | Function f ->
		  if not f.func_used then
		    (
		      f.func_used <- true;
		      check_block f.func_value
		    )
	      | _ -> failwith "internal error"
	  )
    | If(e, s1, s2) ->
	check_expr e;
	check_instr s1.stt_value;
	check_instr s2.stt_value
    | While(e, s) ->
	check_expr e;
	check_instr s.stt_value
    | Compound ss ->
	List.iter (fun s -> check_instr s.stt_value) ss
    | Assign(e1, e2) ->
	check_expr e1;
	check_expr e2
    | Delete e
    | Return e ->
	check_expr e
    | Noop -> ()
  and check_block block =
    List.iter (fun s -> check_instr s.stt_value) block.block_stts
  in
    check_block (snd program)

(**
  Add the definitions of the our external glibb procedures.
*)
let define_glibb program =
  (** @param types type, by reference
  *)
  let new_func_stub name rett types =
    (name, new_function name
       (List.map (fun (t, var) ->
		    {
		      def_name = "";
		      def_type = t;
		      def_nature = Varb;
		      def_value = Variable;
		      def_pos = (-1, -1);
		      def_coord = [];
		      def_asm = Asm_none;
		      def_pass_by_ref = var;
		    }
		 ) types) rett
       {
	 block_env = create_environment ();
	 block_stts = [];
       }
       (-1, -1)
       true)
  in
  let glibb_env =
    [
      new_func_stub "writeInt" Unit [Integer, false];
      new_func_stub "writeBool" Unit [Boolean, false];
      new_func_stub "readInt" Integer [];
    ]
  in
    (snd program).block_env <- concat_env false glibb_env (snd program).block_env

(**
  Check that the program is syntaxically and lexically correct.
*)
let check program =
  define_glibb program;
  check_block (snd program) [] Unit;
  opt_useless program

(**
  Get a human readable representation of an expression.
*)
let rec string_of_expression expr =
  "(" ^ (string_of_type expr.expr_type) ^ ") " ^
  match expr.expr_val with
    | Lt(e1, e2) -> string_of_expression e1 ^ " < " ^ string_of_expression e2
    | Le(e1, e2) -> string_of_expression e1 ^ " <= " ^ string_of_expression e2
    | Eq(e1, e2) -> string_of_expression e1 ^ " = " ^ string_of_expression e2
    | Neq(e1, e2) -> string_of_expression e1 ^ " <> " ^ string_of_expression e2
    | Plus(e1, e2) -> string_of_expression e1 ^ " + " ^ string_of_expression e2
    | Minus(e1, e2) -> string_of_expression e1 ^ " - " ^ string_of_expression e2
    | Mult(e1, e2) -> string_of_expression e1 ^ " * " ^ string_of_expression e2
    | Div(e1, e2) -> string_of_expression e1 ^ " / " ^ string_of_expression e2
    | Mod(e1, e2) -> string_of_expression e1 ^ " mod " ^ string_of_expression e2
    | And(e1, e2) -> string_of_expression e1 ^ " and " ^ string_of_expression e2
    | Or(e1, e2) -> string_of_expression e1 ^ " or " ^ string_of_expression e2
    | Not e -> "not " ^ string_of_expression e
    | Uminus e -> "-" ^ string_of_expression e
    | Int i -> string_of_int i
    | Bool b -> if b then "true" else "false"
    | Var(v, []) -> v.var_name
    | Var(v, el) -> v.var_name ^ "(" ^ String.concat " " (List.map string_of_expression el) ^ ")"
    | Arr(e1, e2) -> string_of_expression e1 ^ "[" ^ (string_of_expression e2) ^ "]"
    | Rec(e, f) -> string_of_expression e ^ "." ^ f
    | Address_of(e) -> "&" ^ string_of_expression e
    | Bang(e) -> string_of_expression e ^ "^"
    | New(t) -> "new " ^ string_of_type !t

(**
  Get a nice human-readable representation of a pascal ast.
*)
let pretty_print (prog_name, main) =
  let get_spaces n =
    if n <= 0 then "" else String.make n ' '
  in
  let string_of_var v =
    (string_of_type v.var_type) ^ " " ^ v.var_name
  in
  let string_of_nature = function
    | Varb -> "var"
    | Cst -> "cst"
    | Funct -> "fun"
    | Unknown_nature -> "???"
  in
  let rec string_of_statement d st =
    get_spaces d ^
    (
      match st.stt_value with
	| Assign(v, e) ->
	    string_of_expression v ^ " := " ^ string_of_expression e
	| Proc(v, el) ->
	    v.var_name ^ " (" ^ String.concat " " (List.map string_of_expression el) ^ ")"
	| Compound stts ->
	    string_of_statements d stts
	| If (e, s1, s2) ->
	    "if " ^ string_of_expression e ^ " then\n" ^ string_of_statement (d + 2) s1 ^ get_spaces d ^ "else\n" ^ string_of_statement (d + 2) s2
	| Delete e ->
	    "delete " ^ string_of_expression e
	| Return e ->
	    "return " ^ string_of_expression e
	| While (e, s) ->
	    "while " ^ string_of_expression e ^ "\n" ^ string_of_statement (d + 2) s
	| Noop -> "NOP"
    ) ^ "\n"
  and string_of_statements d stts =
    String.concat "" (List.map (string_of_statement (d + 4)) stts)
  in
  let rec string_of_value d = function
    | Constant c -> string_of_expression c
    | Variable -> "<var>"
    | Function f ->
	"<fun>\n" ^ string_of_block (d + 2) f.func_value
    | Type t -> "<type : " ^ (string_of_type t) ^ ">"
  and string_of_block d block =
    (fold_env
       (fun n v s ->
	  s ^
	  get_spaces d ^ (string_of_nature v.def_nature) ^ " " ^ (string_of_type v.def_type) ^ " " ^ v.def_name ^ " = " ^ (string_of_value d v.def_value) ^ "\n"
       ) "" block.block_env
    ) ^ (string_of_statements d block.block_stts)
  in
    prog_name ^ "\n" ^ (String.make (String.length prog_name) '=') ^ "\n\n" ^ (string_of_block 0 main)


(** {2 Conversions to [Asm_ir].} *)

(**
  Is an expression a constant?
*)
(* TODO: do not duplicate code like that *)
let is_constant e =
  let rec get_expr_nature e =
    let get_var_def v =
      match v.var_def with
	| Some d -> d
	| None -> raise (Parse_error("internal error (unresolved variable definition)", v.var_pos))
    in
    let is_a_constant v =
      match (get_var_def v).def_value with
	| Constant _ -> true
	| _ -> false
    in
    let score_of_nature = function
      | Cst -> 0
      | Varb -> 1
      | Funct -> 2
      | Unknown_nature -> failwith "trying to get the nature of a not yet typed expression"
    in
    let nature_of_score = function
      | 0 -> Cst
      | 1 -> Varb
      | 2 -> Funct
      | _ -> failwith "internal error"
    in
      match e.expr_val with
	| Lt(e1, e2)
	| Le(e1, e2)
	| Eq(e1, e2)
	| Neq(e1, e2)
	| Plus(e1, e2)
	| Minus(e1, e2)
	| Mult(e1, e2)
	| Div(e1, e2)
	| Mod(e1, e2)
	| And(e1, e2)
	| Or(e1, e2) ->
	    nature_of_score (max (score_of_nature (get_expr_nature e1)) (score_of_nature (get_expr_nature e2)))
	| Uminus e
	| Not e -> get_expr_nature e
	| Int _
	| Bool _ -> Cst
	| Var(v, []) when is_a_constant v -> Cst
	| Var(v, _) -> Varb
	| New _
	| Rec _
	| Arr _ -> Varb (* first approximation *)
	| Address_of _
	| Bang _ -> Varb
  in
    match get_expr_nature e with
      | Cst -> true
      | _ -> false

(**
  Get an asm-like AST corresponding to a pascal-like AST.

  @return a list of procedures incidently defined and the [Asm_ir.procedure] corresponding to the procedure.
*)
let rec asm_of_proc name proc proc_type lex_pos =
  (** Lexical depth. *)
  let lex_depth = List.length lex_pos in

  (** Counter to give a number to procedures. *)
  let pos_in_proc = ref (-1) in
    
  (** Generate a position for a new procedure. *)
  let get_lex_pos () =
    incr pos_in_proc;
    !pos_in_proc :: lex_pos
  in

  (** Create a new variable definition.
    @param t type *)
  let new_var_def t =
    Asm_ir.new_var_def (asm_type t) lex_depth
  in

  (**
    Get the expression corresponding to the address of an expression.
  *)
  let addr_of_expr e =
    {
      expr_val = Address_of e;
      expr_type = Pointer e.expr_type;
      expr_pos = e.expr_pos;
    }
  in

  (**
    Get the nature of a variable.
  *)
  let get_var_nature v =
    if v.var_nature = Unknown_nature then
      v.var_nature <- (get_some v.var_def).def_nature;
    v.var_nature
  in

  (** Handle the prelimnary definitions of the procedure.
    @return a list of defined variables *)
  let handle_defs defs =
    let proc_defs = ref [] in
    let rec handle_def d =
      match d.def_value with
	| Constant c ->
	    (* TODO: this constant might have already been defined higher in the lexical tree *)
	    let cd = new_var_def c.expr_type in
	      cd.Asm_ir.vd_value <- Some (eval_expr c);
	      d.def_asm <- Asm_var cd; Some (Asm_var cd)
	| Variable ->
	    (
	      match d.def_asm with
		| Asm_none ->
		    let v = new_var_def d.def_type in
		      d.def_asm <- Asm_var v; Some (Asm_var v)

		(** It is already defined when it's a variable defined in a procedure. *)
		| def ->
		    Some def
	    )
	| Function f ->
	    if f.func_external || (not f.func_used && !Params.remove_useless_functions) then
	      None
	    else
	      let pd, pasm = asm_of_proc d.def_name f d.def_type (get_lex_pos ()) in
		d.def_asm <- Asm_proc pasm;
		proc_defs := pd @ [pasm] @ !proc_defs;
		Some (Asm_proc pasm)
	| Type _ ->
	    None
    in
    let vars =
      List.fold_left
	(
	  fun l d ->
	    match handle_def d with
	      | Some (Asm_var v) -> v :: l
	      | _ -> l
	) [] defs
    in
      (List.rev vars), !proc_defs
  in

  (** Get a list of asm variables defined in a base block and the asm IR of a base block. *)
  let asm_of_block block =
    let var_with_def d =
      Asm_ir.new_var d
    in

    (** Use a new integer variable. *)
    let fresh_var t =
      let vd = new_var_def t
      in
	vd, Asm_ir.new_var vd
    in
    let get_var_def v =
      match v.var_def with
	| Some d -> d
	| _ -> raise (Asm_error ("internal error in get_var_def (unresolved variable definition)"))
    in
    let get_var_asm v =
      match (get_var_def v).def_asm with
	| Asm_var v -> v
	| _ -> raise (Asm_error ("internal error in get_var_asm (with variable " ^ v.var_name ^ ")"))
    in

    (** Get the flag corresponding to a boolean operation. *)
    let get_expr_flag e =
      match e.expr_val with
	| Lt(_, _) ->
	    {
	      Asm_ir.flag_not = false;
	      Asm_ir.flag_flag = Asm_ir.Lflag;
	      Asm_ir.flag_eq = false;
	    }
	| Le(_, _) ->
	    {
	      Asm_ir.flag_not = false;
	      Asm_ir.flag_flag = Asm_ir.Lflag;
	      Asm_ir.flag_eq = true;
	    }
	| Eq(_, _) ->
	    {
	      Asm_ir.flag_not = false;
	      Asm_ir.flag_flag = Asm_ir.Eflag;
	      Asm_ir.flag_eq = false;
	    }
	| Neq(_, _) ->
	    {
	      Asm_ir.flag_not = true;
	      Asm_ir.flag_flag = Asm_ir.Eflag;
	      Asm_ir.flag_eq = false;
	    }
	| New _
	| Bang _
	| Address_of _
	| Rec _
	| Arr _
	| Var _
	| Bool _
	| Int _
	| Not _
	| And _
	| Mod _
	| Div _
	| Mult _
	| Or _
	| Minus _
	| Plus _
	| Uminus _ -> raise (Asm_error ("internal error in get_expr_flag (expression is not a comparison)"))
    in

    (**
      Get the asm IR of an expression.
      @returns the created definitions, the asm instructions and the variable where the result is avilable.
    *)
    let rec get_expr_asm e =
      let vd, res = fresh_var (e.expr_type) in
	if is_constant e then
	  (
	    vd.Asm_ir.vd_value <- Some (eval_expr e);
	    [vd], [], res
	  )
	else
	  match e.expr_val with
	    | Plus(e1, e2) ->
		let vd1, i1, v1 = get_expr_asm e1 in
		let vd2, i2, v2 = get_expr_asm e2 in
		  vd :: vd1 @ vd2, i1 @ i2 @ [Asm_ir.Add_i(res, v1, v2)], Asm_ir.copy_var res
	    | Minus(e1, e2) ->
		let vd1, i1, v1 = get_expr_asm e1 in
		let vd2, i2, v2 = get_expr_asm e2 in
		  vd :: vd1 @ vd2, i1 @ i2 @ [Asm_ir.Sub_i(res, v1, v2)], Asm_ir.copy_var res
	    | Mult(e1, e2) ->
		let vd1, i1, v1 = get_expr_asm e1 in
		let vd2, i2, v2 = get_expr_asm e2 in
		  vd :: vd1 @ vd2, i1 @ i2 @ [Asm_ir.Mult_i(res, v1, v2)], Asm_ir.copy_var res
	    | Div(e1, e2) ->
		let vd1, i1, v1 = get_expr_asm e1 in
		let vd2, i2, v2 = get_expr_asm e2 in
		  vd :: vd1 @ vd2, i1 @ i2 @ [Asm_ir.Div_i(res, v1, v2)], Asm_ir.copy_var res
	    | Mod(e1, e2) ->
		let vd1, i1, v1 = get_expr_asm e1 in
		let vd2, i2, v2 = get_expr_asm e2 in
		  vd :: vd1 @ vd2, i1 @ i2 @ [Asm_ir.Mod_i(res, v1, v2)], Asm_ir.copy_var res
	    | Uminus(e1) ->
		let vd1, i1, v1 = get_expr_asm e1 in
		  vd :: vd1, i1 @ [Asm_ir.Neg_i(res, v1)], Asm_ir.copy_var res
	    | Eq(e1, e2)
	    | Neq(e1, e2)
	    | Lt(e1, e2)
	    | Le(e1, e2) ->
		let vd1, i1, v1 = get_expr_asm e1 in
		let vd2, i2, v2 = get_expr_asm e2 in
		  vd :: vd1 @ vd2, i1 @ i2 @ [Asm_ir.Cmp_i(res, v1, get_expr_flag e, v2)], Asm_ir.copy_var res
	    | And(e1, e2) ->
		let vd1, i1, v1 = get_expr_asm e1 in
		let vd2, i2, v2 = get_expr_asm e2 in
		  vd :: vd1 @ vd2, i1 @ i2 @ [Asm_ir.And(res, v1, v2)], Asm_ir.copy_var res
	    | Or(e1, e2) ->
		let vd1, i1, v1 = get_expr_asm e1 in
		let vd2, i2, v2 = get_expr_asm e2 in
		  vd :: vd1 @ vd2, i1 @ i2 @ [Asm_ir.Or(res, v1, v2)], Asm_ir.copy_var res
	    | Not e1 ->
		let vd1, i1, v1 = get_expr_asm e1 in
		  vd :: vd1, i1 @ [Asm_ir.Not(res, v1)], Asm_ir.copy_var res
	    | Int i ->
		(* TODO: check that it fits in an asm instruction *)
		(* TODO: use Asm_ir.new_const_int_var? *)
		vd.Asm_ir.vd_value <- Some (Asm_ir.Int i); [vd], [], res
	    | Bool b ->
		vd.Asm_ir.vd_value <- Some (Asm_ir.Bool b); [vd], [], res
	    | New t ->
		[vd], [Asm_ir.New(res,sizeof !t)], Asm_ir.copy_var res
	    | Arr(a, n) ->
		let vda, ia, va = get_expr_asm a in
		let vdn, i_n, vn = get_expr_asm n in
		let t, n0 =
		  match a.expr_type with
		    | Array(t, n0, _) -> t, n0
		    | _ -> raise (Asm_error ("internal error in get_expr_asm (using a variable which is not an array as an array)"))
		in
		let n0 = eval_expr_int n0 in
		  vd.Asm_ir.vd_pos := Asm_ir.Pos_arr(va, vn, n0, sizeof t);
		  vd :: vda @ vdn, ia @ i_n, res
	    | Rec(r, f) ->
		let vdr, ir, vr = get_expr_asm r in
		let k =
		  let rec get_k = function
		    | (field, t) :: _ when f = field -> 0
		    | (_, t) :: tl -> (sizeof t) + get_k tl
		    | [] -> failwith "error" (* TODO: error, field not found *)
		  in
		    match r.expr_type with
		      | Record r -> get_k r
		      | _ -> failwith "error" (* TODO: error, this is not a record! *)
		in
		  vd.Asm_ir.vd_pos := Asm_ir.Pos_rec(vr, k);
		  vd :: vdr, ir, res
	    | Address_of e1 ->
		let vd1, i1, v1 = get_expr_asm e1 in
		  Asm_ir.set_var_ots v1;
		  (* TODO: the spill var is not enough !!!! if the value is modified, the associated pointer could be not changed because of registers !!! *)
		  vd :: vd1, i1 @ [Asm_ir.Spill_var(Asm_ir.copy_var v1); Asm_ir.Address_of(res, v1)], Asm_ir.copy_var res
	    | Bang e1 ->
		let vd1, i1, v1 = get_expr_asm e1 in
		let vp = Asm_ir.new_var v1.Asm_ir.var_def in
		  vp.Asm_ir.var_def.Asm_ir.vd_lex_depth := lex_depth;
		  vd.Asm_ir.vd_pos := Asm_ir.Pos_pointed vp;
		  vd :: vd1, i1, res
	    | Var(v, []) when get_var_nature v <> Funct ->
		res.Asm_ir.var_def <- get_var_asm v; [], [], res
	    | Var(v, a) ->
		(* TODO: share this code with Proc *)
		let d, i, s = (** Instructions and return variable. *)
		  List.fold_left2
		    (
		      fun (d, i, a) arg e ->
			let d', i', a' =
			  get_expr_asm
			    (
			      if arg.def_pass_by_ref then
				addr_of_expr e
			      else
				e
			    )
			in
			  d' @ d, i' :: i, a' :: a
		    ) ([], [], [])
		    (
		      match (get_var_def v).def_value with
			| Function f -> f.func_args
			| _ -> raise (Asm_error ("internal error in get_expr_asm (using a variable which is not an function as a function)"))
		    ) a
		in
		let i = List.concat (List.rev i) in
		let p =
		  match (get_var_def v).def_value with
		    | Function f -> f
		    | _ -> raise (Parse_error("internal error, procedure is not a procedure", v.var_pos)) (* TODO: test this *)
		in
		  vd :: d, i @ [Asm_ir.Fun(res, p.func_asm_def, s)], Asm_ir.copy_var res
		    
    in
    let new_bb i mi =
      {
	Asm_ir.bb_block = i;
	Asm_ir.bb_link = mi;
      }
    in
    let add_to_bb bb i =
      bb.Asm_ir.bb_block <- i @ bb.Asm_ir.bb_block
    in
    let append_to_bb bb i =
      bb.Asm_ir.bb_block <- bb.Asm_ir.bb_block @ i
    in
    let get_stts_list stt =
      match stt.stt_value with
	| Compound c -> c
	| _ -> [stt]
    in

    (** Get a base block corresponding to a list of statements (the first argument of the returned couple is a list of variable definitions needed for temporary variables). *)
    (* TODO: tail-rec? *)
    let rec asm_of_stts statements =
      match statements with
	| [] -> [], new_bb [] Asm_ir.End
	| stt :: stts ->
	    let bbd, bbs = asm_of_stts stts in
	      match stt.stt_asm with
		  (* TODO: is this useful? *)
		| Some(d, mi) -> d @ bbd, new_bb [] mi
		| None ->
		    match stt.stt_value with
		      | Assign(n, e) ->
			  let nd, ni, ns = get_expr_asm n in
			  let d, i, s = get_expr_asm e in
			    add_to_bb bbs (i @ ni @ [Asm_ir.Mov_i(ns, s)]);
			    d @ nd @ bbd, bbs
		      | Proc(v, a) ->
			  let defs = ref [] in (** Defined variables. *)
			  let i, s = (** Instructions and return variable. *)
			    List.fold_left2
			      (
				fun (i, a) arg e ->
				  let d', i', a' =
				    get_expr_asm
				      (
					if arg.def_pass_by_ref then
					  addr_of_expr e
					else
					  e
				      )
				  in
				    defs := d' @ !defs;
				    i' :: i, a' :: a
			      ) ([], [])
			      (
				match (get_var_def v).def_value with
				  | Function f -> f.func_args
				  | _ -> raise (Asm_error ("internal error in asm_of_stts (using a variable which is not an procedure as a procedure)"))
			      ) a
			  in
			  let i = List.concat (List.rev i) in
			  let p =
			    match (get_var_def v).def_value with
			      | Function f -> f
			      | _ -> raise (Parse_error("internal error, procedure is not a procedure", v.var_pos)) (* TODO: test this *)
			  in
			    (* TODO: globally unique name *)
			    add_to_bb bbs (i @ [Asm_ir.Proc(p.func_asm_def, s)]); !defs @ bbd, bbs
		      | If(e, s1, s2) ->
			  let de, ae, re = get_expr_asm e
			  and ds1, as1 = asm_of_stts (get_stts_list s1)
			  and ds2, as2 = asm_of_stts (get_stts_list s2)
			  in
			  let bbr = new_bb [] (Asm_ir.If(re, as1, as2, bbs)) in
			    add_to_bb bbr ae;
			    de @ ds1 @ ds2 @ bbd, bbr
		      | While(e, s1) ->
			  let de, ae, re = get_expr_asm e
			  and ds1, as1 = asm_of_stts (get_stts_list s1)
			  in
			  let bbr = new_bb [] (Asm_ir.While(re, new_bb ae Asm_ir.End, as1, bbs)) in
			    de @ ds1 @ bbd, bbr
		      | Delete e ->
			  let d, i, s = get_expr_asm e in
			    add_to_bb bbs (i @ [Asm_ir.Delete s]);
			    d @ bbd, bbs
		      | Return e ->
			  let d, i, s = get_expr_asm e
			  in
			    add_to_bb bbs (i @ [Asm_ir.Return s]);
			    d @ bbd, bbs
			      (* TODO: don't compute bbs in this case *)
		      | Compound cstts ->
			  asm_of_stts (cstts @ stts)
		      | Noop -> bbd, bbs
			  
    in
      asm_of_stts block.block_stts
  in

  (** Remove the arguments from the environment of a block. *)
  let rec remove_args args env =
    match args, env with
      | _ :: a, _ :: e -> remove_args a e
      | [], e -> e
      | _, [] -> raise (Asm_error "internal error in remove_args")
  in
  let vars, pd = handle_defs (List.map (snd) proc.func_value.block_env) in
  let vars = (remove_args proc.func_args vars) in
  let nvars, asm = asm_of_block proc.func_value in
  let lex_depth = List.length lex_pos in
    List.iter
      (fun a ->
	 match a.def_asm with
	   | Asm_var vd ->
	       vd.Asm_ir.vd_lex_depth := lex_depth
	   | _ -> raise (Asm_error ("internal error in asm_of_proc (arguments of procedure are not variables)"))
      ) proc.func_args;
    proc.func_asm_def.Asm_ir.proc_lex_depth <- lex_depth;
    proc.func_asm_def.Asm_ir.proc_vars <- nvars @ vars;
    proc.func_asm_def.Asm_ir.proc_value <- asm;
    proc.func_asm_def.Asm_ir.proc_is_fun <- proc_type <> Unit;
    pd, proc.func_asm_def

(**
  Get the asm IR of a program.
*)
let asm_of_prog prog =
  let name, block = prog in
  let procs, main = asm_of_proc "main"
		      {
			func_args = [];
			func_value = block;
			func_asm_def = Asm_ir.proc_stub "main" [] 0;
			func_used = true;
			func_external = false;
		      } Unit []
  in
    {
      Asm_ir.prog_source_name = "";
      Asm_ir.prog_procs = procs @ [main];
      Asm_ir.prog_max_depth = List.fold_left (fun m p -> max m p.Asm_ir.proc_lex_depth) 0 procs;
    }
