(*
  Bifton - A Pascal- compiler.

  Copyright (C) 2004 Samuel Mimram & Julien Cristau.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(**
  sparc backend.

  @author Samuel Mimram, Julien Cristau
*)

(* $Id: sparc.ml,v 1.11 2004/05/09 20:02:08 julio Exp $ *)

(*
  TODO:
  - global tables: display, table of constants (?)
*)

open Asm_ir

let string_of_char c =
  String.make 1 c

let rec list_assoc_id x = function
  | (a, b) :: t when x.vd_ident = a.vd_ident -> b
  | _ :: t -> list_assoc_id x t
  | [] -> raise Not_found

(* TODO: improve *)
let rec list_mem_assoc_id x l =
  try
    ignore (list_assoc_id x l);
    true
  with
    | Not_found -> false

let rec list_assoc' x = function
  | (a, b) :: t when x = b -> a
  | _ :: t -> list_assoc' x t
  | [] -> raise Not_found

let rec list_mem_assoc' x = function
  | (a, b) :: t when x = b -> true
  | _ :: t -> list_mem_assoc' x t
  | [] -> false

let list_itern f l =
  let rec aux n = function
    | h :: t -> f n h; aux (n + 1) t
    | [] -> ()
  in
    aux 0 l

let sp = Reg_int 14
let fp = Reg_int 30

(* dirty hack : do not allocate a variable in these registers 
0 is always null, 1 is reserved for temporary use while spilling,
14 is %sp, 15 is the address of the call instruction, and 30 is %fp *)
let reserved_registers = [ 0; 1; 14; 15; 30]

let get_reg_var r =
  let vd = new_var_def Pointer (-1) in
  let v = new_var vd in
    v.var_reg <- Some r; v

(**
  Get a variable containing the address of the top of the stack.
*)
let get_stack_pos_var () =
  get_reg_var sp

(** Total number of available registers (warning: register numbers are zero-based). *)
let nb_registers = 29

(**
  Do the registers allocation. This is done base block by base block (all registers should be available at the beginning of each base block and all variables needed again should be spilled at the end of each base block). %edx is reserved for loading values passed by reference.
*)
let allocate_registers base_block lex_depth =
  (**
    @param av list of available variables (list of definition and associated register)
    @param instr the instruction code should be generated for
    @param needed_defs definition of variables whose values are used in the rest of the block
    @returns a list of available variables after the instruction and a rewrite of the code *)
  (* TODO: really compute and use needed_defs *)
  let rec alloc_instr av instr needed_defs =
    let av = ref av in
      (** Declare that a variable is available in a register. If some other variable was previously declared as available in the same register, it won't be anymore. *)
    let set_available v r =
      if list_mem_assoc_id v !av then failwith "a variable has two locations in memory!";
      (* if list_mem_assoc' r !av then failwith "two variables are in the same memory location!"; *)
      av := (v, r) :: (List.filter (fun (_, r') -> r' <> r) !av)
    in
      (** Warning [before] and [after] are reversed at the end of the function! *)
    let before = ref []
    and after = ref [] in
    let do_before a =
      before := a :: !before in
    let do_after a = 
      after := a:: !after
    in

    (* TODO: remove *)
    let print_av av =
      do_before
	(Code
	   ((List.fold_left
	       (fun s ->
		  (function
		     | (vd, Reg_int i) -> s ^ " " ^ (string_of_int i) ^ "," ^ (string_of_int vd.vd_ident)
		     | _ -> s ^ " ?"
		  )) "av:" av) ^ "\n")
	)
    in
      (* let do_before a =
	 do_before a; print_av !av;
	 in *)

    let during =
      (** Mark a register as unused. *)
      let free_reg r =
	av := List.filter (fun (_, r') -> r' <> r) !av
      in
      let rec do_spill r vd =
	vd.vd_on_the_stack := true;
	match get_vd_type vd with
	  | Other _ -> ()
	  | _ -> 
	      do_before (Spill(get_vd_lex vd <> lex_depth, r, vd))
      and do_unspill v =
	match get_var_type v with
	  | Other _ -> 
	      do_before (Address_of(v,v))
	  | _ ->
	      do_before (Unspill (get_var_lex v <> lex_depth, v))

      (** Allocate a variable in a specific register. *)
      and alloc_in_reg v r occ =
	(** Warning: v.var_reg should be set for flags! *)
	match v.var_reg with
	  | Some ((Reg_flag f) as rf) ->
	      v.var_reg <- Some r;
	      do_before (Set_var(f, v));
	      free_reg rf;
	      set_available v.var_def r; r
	  | Some r -> r
	  | None ->
	      (
		match v.var_def.vd_value with
		  | Some _ ->
		      v.var_reg <- Some r;
		      let v' = copy_var v in
			v'.var_reg <- None;
			do_before(Mov_i(v,v'));
			set_available v.var_def r; r
		  | None ->
		      v.var_reg <- Some r;
		      do_unspill (copy_var v);
		      set_available v.var_def r; r
	      )

      and get_avail_reg occ =
	let rec find_free_int n =
	  if n >= nb_registers then raise Not_found
	  else if List.mem (Reg_int n) occ then
	    find_free_int (n + 1)
	  else if list_mem_assoc' (Reg_int n) !av then
	    find_free_int (n + 1)
	  else if List.mem n reserved_registers then
	    find_free_int (n + 1)
	  else
	    Reg_int n
	in
	  (* TODO: spill the one which will not be used during the most long time *)
	let rec find_to_spill_int n =
	  if n >= nb_registers then raise (Asm_error "no registers left during assembly generation")
	  else if List.mem (Reg_int n) occ then
	    find_to_spill_int (n + 1)
	  else
	    Reg_int n
	in
	  try
	    find_free_int 1
	  with
	    | Not_found ->
		let r = find_to_spill_int 1 in
		let replaced_vd = list_assoc' r !av in
		  spill_free true r replaced_vd occ; r

      (** Allocate an integer register to the variable [v].
	@param occ is a list of already taken integer registers *)
      and alloc_reg v occ =
	(* TODO: should this be there or in check_avail? *)
	let occ = 
	  match get_var_pos v with
	    | Pos_reg r -> r :: occ
	    | Pos_arr(a, v, _, _) -> get_var_reg a :: get_var_reg v :: occ
	    | Pos_rec(p, _)
	    | Pos_pointed p -> get_var_reg p :: occ
	    | Pos_base _
	    | Pos_stack _
	    | Pos_label _
	    | Pos_none -> occ
	in
	  alloc_in_reg v (get_avail_reg occ) occ

      (* TODO: if it's not needed anymore, don't spill it *)
      (** @param free should we declare the freed variable's register as free?
	@param occ is here to specify needed registers because we might need an extra register during the spill *)
      and spill_free free r vd occ =
	vd.vd_on_the_stack := true;
	(
	  match get_vd_pos vd with
	    | Pos_arr(b,v,_,_) -> 
		assert_avail b occ; assert_avail v occ
	    | Pos_pointed p
	    | Pos_rec(p,_) -> assert_avail p occ
	    | _ -> ()
	);
	(
	  match r with
	    | Reg_int _ ->
		do_spill r vd;
		if free then free_reg r
	    | Reg_flag f ->
		let v = new_var vd in
		  v.var_reg <- Some r;
		  free_reg r;
		  (
		    try
		      v.var_reg <- Some (list_assoc_id v.var_def !av)
		    with
		      | _ -> ignore (alloc_reg v occ)
		  );
		  let r' = get_var_reg v in
		    do_spill r' vd;
		    free_reg r';
		    if free then free_reg r
	);

      (** If the variable is in a register then spill it. Its register is then marked as free. *)
      and spill_var_free free v occ =
	check_avail_flags true v occ (fun l -> List.iter (fun r -> spill_free free r v.var_def occ) l)

      (** Make sure that a specific register is available. *)
      and spill_reg r occ =
	List.iter (fun (vd, rd) -> if r = rd then spill_free true rd vd occ) !av 

      (** If the variable [v] is available then we should execute [f] which will receive [v]'s register as argument else we should execute [g]. *)
      and on_avail v f g =
	try
	  let r = list_assoc_id v.var_def !av in
	    f r
	with
	  | Not_found -> g ()

      (** If variable [v] already was in a register it stays in it. The function [f] is called with the list of used registers. The second argument tells whether it is acceptable or not for it to be stored in flags. *)
      and check_avail_flags flags v occ f =
	let check_pos occ = function
	  | Pos_reg r -> r :: occ
	  | Pos_rec(v, _)
	  | Pos_pointed v -> assert_avail v occ; get_var_reg v :: occ
	  | Pos_arr(a, v, _, _) ->
	      assert_avail a occ;
	      let occ = get_var_reg a :: occ in
		assert_avail v occ;
		get_var_reg v :: occ
	  | Pos_base _ | Pos_stack _ | Pos_label _ | Pos_none -> occ
	in
	  v.var_reg <- None; (* TODO: is that useful? *)
	  on_avail v
	    (fun r ->
	       let r =
		 if flags && !Params.use_flags then r
		 else
		   match r with
		     | Reg_flag _ ->
			 v.var_reg <- Some r;
			 alloc_reg v occ
		     | _ -> r
	       in
		 v.var_reg <- Some r; f [r]
	    )
	    (fun () ->
	       let occ = check_pos occ (get_var_pos v) in
		 if get_var_lex v <> lex_depth || get_var_type v = Boolean then
		   f [alloc_reg v occ]
		 else
		   (
		     if not (is_constant v) then
		       v.var_def.vd_on_the_stack := true;
		     f []
		   )
	    )

      (** If the variable is not yet in a register put it in some register. *)
      and assert_avail v occ =
	check_avail_flags false v occ
	  (fun r ->
	     match r with
	       | [] -> ignore (alloc_reg v occ)
	       | [Reg_flag f] ->
		   ignore (alloc_reg v occ) 
	       | _ -> ()
	  )
      in
      let spill = spill_free true in

      let spill_var = spill_var_free true in

      let spill_all () =
	List.iter (fun (vd, r) ->
		     match r with
		       | Reg_int _ ->
			   vd.vd_on_the_stack := true;
			   spill r vd []
		       | Reg_flag _ -> () (* TODO! *)
		  ) !av
      in
      let check_avail = check_avail_flags false in
      let just_check_avail v occ =
	check_avail v occ (fun _ -> ())
      in
      let used_and_replaced v occ = 
	if !(Params.minimize_spills) then
	  (
	    spill_var_free false v occ;
	    assert_avail v occ;
	    free_reg (get_var_reg v)
	  )
	else
	  (
	    spill_var v occ;
	    assert_avail v occ
	  )
      in
	match instr with
	  | Mov_b(v1, v2)
	  | Not(v1, v2)
	  | Neg_i(v1, v2)
	  | Mov_i(v1, v2) ->
	      assert_avail v2 [];
	      ignore(alloc_reg v1 [get_var_reg v2]);
	      [instr]
	  | Address_of(v1, v2) ->
	      ignore(alloc_reg v1 []);
	      [instr]
	  | New(v1, n) ->
	      set_available v1.var_def (Reg_int 8);
	      [instr]
	  | Mult_i(v1, v2, v3)
	  | And(v1, v2, v3)
	  | Or(v1, v2, v3)
	  | Div_i(v1, v2, v3)
	  | Mod_i(v1, v2, v3)
	  | Add_i(v1, v2, v3)
	  | Sub_i(v1, v2, v3) ->
	      assert_avail v2 [];
	      assert_avail v3 [get_var_reg v2];
	      ignore(alloc_reg v1 [get_var_reg v2; get_var_reg v3]);
	      [instr]
	  | Cmp_i(v1, v2, f, v3) ->
	      assert_avail v2 [];
	      assert_avail v3 [get_var_reg v2];
	      ignore(alloc_reg v1 [get_var_reg v2; get_var_reg v3]);
	      do_after(Set_var (f,v1));
	      [instr]
	  | Proc(p, l) ->
	      spill_all();
	      let argv = Array.of_list l in
		for i = 0 to min 5 ((Array.length argv)-1) do
		  let v = argv.(i) in
		    match get_var_type v with
		      | Other _ ->
			  assert_avail v [];
			  do_before(Push_o(v, sizeof(get_var_type v)))
		      | _ ->
			  let a = get_reg_var(Reg_int (i+8)) in
			    assert_avail argv.(i) [];
			    do_before(Mov_i (a,argv.(i)))
		done;
		for i = 6 to Array.length argv - 1 do
		  let v = argv.(i) in
		    match get_var_type v with
		      | Other _ -> 
			  assert_avail v [];
			  do_before(Push_o(v, sizeof(get_var_type v)))
		      | _ ->
			  do_before(Push argv.(i))
		done;
		[Call p.proc_name; Nop]
	  | Fun(x,p,l) ->
	      spill_all();
	      let argv = Array.of_list l in
		for i = 0 to min 5 ((Array.length argv)-1) do
		  let v = argv.(i) in
		    match get_var_type v with
		      | Other _ ->
			  assert_avail v [];
			  do_before(Push_o(v, sizeof(get_var_type v)))
		      | _ ->
			  let a = get_reg_var(Reg_int (i+8)) in
			    assert_avail argv.(i) [];
			    do_before(Mov_i (a,argv.(i)))
		done;
		for i = 6 to Array.length argv - 1 do
		  let v = argv.(i) in
		    match get_var_type v with
		      | Other _ -> 
			  assert_avail v [];
			  do_before(Push_o(v, sizeof(get_var_type v)))
		      | _ ->
			  do_before(Push argv.(i))
		done;
		set_available x.var_def (Reg_int 8);
		[Call p.proc_name; Nop]
	  | Delete _
	  | Return _ 
	  | Nop -> [instr]
	  | Assert_avail v -> assert_avail v []; []
	  | Check_avail_with_flags v -> check_avail_flags false v [] (fun _ -> ()); []
	  | Spill_var v ->
	      spill_var v []; []
	  | Spill_all ->
	      spill_all (); []
	  | Code _
	  | Set_var _
	  | Unspill _
	  | Spill _
	  | Push_o _
	  | Push _
	  | Call _
	  | Check_avail _
	  | Read_flag _ -> raise (Asm_error "internal error in alloc_instr (encountered an instruction which should not have been generated)")
    in
      !av, (List.rev !before) @ during @ (List.rev !after)
  in
    (** Allocate registers for a list of instructions.
      @returns a list of available variables after the instruction and a rewrite of the code *)
  let rec alloc_instrs av nd = function
    | h :: t ->
	let av', i' = alloc_instr av h [] (*used_var_defs_in_instrs t*) in
	let av'', i'' = alloc_instrs av' nd t in
	  av'', i' @ i''
    | [] -> av, []
  in
  let rec alloc_base_block av base_block =
    (** Assembly needed by the following base blocks. *)
    let needed =
      match base_block.bb_link with
	| If(v, _, _, _) -> [Check_avail_with_flags v]
	| While _ 
	| End -> []
    in
    let av, instrs = alloc_instrs av [] (base_block.bb_block @ [Spill_all] @ needed) in
    let instrs = ref instrs in
      (
	match base_block.bb_link with
	  | If(v, b1, b2, b3) ->
	      let av1 = alloc_base_block av b1 in
	      let av2 = alloc_base_block av b2 in
		ignore (alloc_base_block
			  [] (* TODO: we could use the "intersection" of av1 and av2 *)
			  b3)
	  | While(v, be, b1, b2) ->
	      be.bb_block <- be.bb_block @ [Check_avail_with_flags v];
	      let ave = alloc_base_block av be in
	      let av1 = alloc_base_block ave b1 in (* TODO: is ave correct? *)
		ignore (alloc_base_block av1 b2)
	  | End -> (* (** Let's spill everything. *)
		      instrs := !instrs @ (List.fold_left (fun l (vd, r) ->
		      match r with
		      | Reg_int _ ->
		      vd.vd_on_the_stack := true;
		      Spill(get_ vd <> lex_depth, r, vd) :: l
		      | Reg_flag _ -> l (* TODO! *)
		      ) [] av) *) ()
      );
      base_block.bb_block <- !instrs; av
  in
    ignore (alloc_base_block [] base_block)

(**
  Generate the target asm.
*)

let generate_code prog =
  let ret = ref "" in
  let next_label = ref 0 in
  let get_fresh_label () =
    incr next_label;
    "L" ^ (string_of_int !next_label)
  in
  let code_of_proc p =
    let string_of_register = function
      | Reg_int n -> if n<0 || n>31 then 
	  raise (Asm_error "trying to use a not supported register !")
	else (
	  if n = 30 then "%fp" else
	    if n = 14 then "%sp" else
	      if n<8 then "%g" ^ (string_of_int n)
	      else if n<16 then "%o" ^ (string_of_int (n-8))
	      else if n<24 then "%l" ^ (string_of_int (n-16))
	      else "%i" ^ (string_of_int (n-24))
	)
      | Reg_flag _ -> raise(Asm_error "Trying to access a flag register !")
    in
    let string_of_flag f = 
      match f.flag_flag with
	| Aflag -> if f.flag_eq then "cc" else "gu"
	| Bflag -> if f.flag_eq then "leu" else "cs"
	| Cflag -> if f.flag_not then "cc" else "cs"
	| Eflag -> if f.flag_not then "ne" else "e"
	| Gflag -> if f.flag_eq then "ge" else "g"
	| Lflag -> if f.flag_eq then "l" else "le"
	| Oflag -> if f.flag_not then "vc" else "vs"
	| Sflag -> if f.flag_not then "neg" else "pos"
	| Zflag 
	| Pflag
	| Vflag _ -> raise (Asm_error "internal error in string_of_flag")
    in
    let jump_of_flag f = "b" ^ (string_of_flag f) in
(*
    let string_of_stack_pos = function
      | Pos_base n -> "%fp" ^ if n>=0 then "+" else "" ^ (string_of_int n) 
      | Pos_stack n -> "%sp" ^ if n>=0 then "+" else "" ^ (string_of_int n)
      | Pos_label n ->  n
      | Pos_none -> "(%fuck fuck)" (* raise (Asm_error "trying to get the position of variable which is not on the stack") *)
      | _ -> failwith "ooops"
  in
  let code_of_proc p =
    (** @return register's name (or maybe a constant value). *)
    (* TODO: do not take depth in accound for Pos_stack and Pos_label *)
    let code_of_var v =
      match v.var_reg with
	| Some r -> string_of_register r
	| None ->
	    (
	      match v.var_def.vd_value with
		| Some(Int i) -> string_of_int i
		| _ -> string_of_stack_pos (get_var_pos v)
	    )
    in
*)
      (** Generate the code for accessing to the ARP of an other procedure. *)
    let get_display lex_depth r =
Instr_sparc.set ("(DISPLAY" ^ (if lex_depth = 0 then "" else "+" ^ string_of_int (4 * lex_depth)) ^ ")") (string_of_register r) ^
      Instr_sparc.load ("[" ^ string_of_register r ^ "]") (string_of_register r)
    in
    let code_of_instr instr =
      let before = ref "" in
      let do_before s =
	before := !before ^ s
      in
      let rec string_of_stack_pos base pos =
	let rec get_reg_n = function
	  | Pos_base n -> string_of_register base, n
	  | Pos_stack n -> "%sp", n
	  | Pos_pointed v -> code_of_var v, 0
	  | Pos_rec _
	  | Pos_arr _
	  | Pos_label _
	  | Pos_reg _
	  | Pos_none -> raise (Asm_error "internal error in get_reg_n")
	in
	  match pos with
	    | Pos_label n -> n
	    | Pos_rec(r, k) ->
		(string_of_register (get_var_reg r)) ^ 
		(if k=0 then "" else ((if k>0 then "+" else "" ) ^ (string_of_int k)))
	    | Pos_arr(a, v, n0, s) ->
		(
		  let ar = string_of_register (get_var_reg a) in
		    match v.var_def.vd_value with
		      | Some (Int i) ->
			  let k = (i - n0) * s in
			  ar ^ (if k=0 then "" else ((if k>0 then "+" else "") ^ (string_of_int k)))
		      | _ ->
			  let vr = get_var_reg v in
			    do_before (""); ""
		)
	    | Pos_none -> raise (Asm_error "trying to get the position of variable which is not on the stack")
	    | Pos_base _
	    | Pos_stack _
	    | Pos_pointed _ ->
		let r,n = get_reg_n pos in
		  "[" ^ r ^ (if n = 0 then "" else ((if n > 0 then "+" else "") ^ string_of_int n)) ^ "]"
	    | Pos_reg r -> string_of_register r

      and code_of_var_no_reg v =
	match v.var_def.vd_value with
	  | Some(Int i) ->  (string_of_int i)
	  | Some(Bool false) -> "0"
	  | Some(Bool true) -> "1"
	  | _ -> string_of_stack_pos fp (get_var_pos v)
      and code_of_var_reg v =
	match v.var_reg with
	  | Some r -> string_of_register r
	  | None -> code_of_var_no_reg v
      and code_of_var v =
	if is_constant v then
	  code_of_var_no_reg v
	else
	  code_of_var_reg v
      in
	!before ^
	(* TODO: optimisations should be done before, not here! *)
	match instr with
	    (* TODO: only if both args are <> 0 (handle constants) *)
	  | Or (v1, v2, v3) ->
	      let r1 = code_of_var_reg v1 in
	      let r2 = code_of_var v2 in
	      let r3 = code_of_var v3 in
		Instr_sparc.log_or r2 r3 r1
	  | And (v1, v2, v3) ->
	      let r1 = code_of_var_reg v1 in
	      let r2 = code_of_var v2 in
	      let r3 = code_of_var v3 in
		Instr_sparc.log_and r2 r3 r1
	  | Not(v1,v2) ->
	      let r1 = code_of_var_reg v1 in
	      let r2 = code_of_var_reg v2 in
		Instr_sparc.xor r2 "1" r1
	  | Neg_i(v1, v2) ->
	      let r1 = code_of_var_reg v1 in
	      let r2 = code_of_var_reg v2 in
		Instr_sparc.sub "%g0" r2 r1
	  | Add_i(v1, v2, v3) ->
	      let r1 = code_of_var_reg v1 in
	      let r2 = code_of_var v2 in
	      let r3 = code_of_var v3 in
		Instr_sparc.add r2 r3 r1
	  | Sub_i(v1, v2, v3) ->
	      let r1 = code_of_var_reg v1 in
	      let r2 = code_of_var v2 in
	      let r3 = code_of_var v3 in
		Instr_sparc.sub r2 r3 r1
	  | Mult_i(v1, v2, v3) ->
	      let r1 = code_of_var_reg v1 in
	      let r2 = code_of_var v2 in
	      let r3 = code_of_var v3 in
		Instr_sparc.mult r2 r3 r1
	  | Div_i(v1, v2, v3) ->
	      let r1 = code_of_var_reg v1 in 
	      let r2 = code_of_var v2 in
	      let r3 = code_of_var v3 in
		(* clear Y register and 3 nops *)
		Instr_sparc.wry "%g0" "%g0" ^
		Instr_sparc.nop ^ Instr_sparc.nop ^ Instr_sparc.nop ^
		(* compute the division *)
		Instr_sparc.div r2 r3 r1
	  | Cmp_i(v1, v2, _, v3) ->
(*	      let r1 = code_of_var v1 in *)
	      let r2 = code_of_var v2 in
	      let r3 = code_of_var v3 in
		Instr_sparc.cmp r2 r3
	  | Mov_i(v1, v2) ->
	      let r1 = code_of_var_reg v1 in
	      let r2 = code_of_var v2 in
		Instr_sparc.move r2 r1
	  | New(v1, n) ->
	      let r1 = code_of_var v1 in
		Instr_sparc.move (string_of_int n) "%o0" ^ 
		Instr_sparc.call "sbrk" ^ Instr_sparc.nop (* ^
		Instr_sparc.move "%o0" r1 *)
	  | Delete v1 -> ""
	  | Address_of(v1, v2) ->
	      let r1 = code_of_var v1 in
		get_display (get_var_lex v2) (get_var_reg v1) ^
		Instr_sparc.move (string_of_stack_pos (get_var_reg v1) (get_var_pos v2)) r1
	  | Unspill(lex_scope, v) ->
	      if not (get_var_ots v) then ""
	      else
		(
		  let r' = code_of_var v in
		    if lex_scope then
		      match get_var_pos v with
			| Pos_base b ->
			    (
			      match v.var_reg with
				| None -> raise (Asm_error "trying to access a variable in an other lexical scope without being assigned to a register")
				| Some r -> 
				    get_display (get_var_lex v) r ^
				    Instr_sparc.load 
				      ("[" ^ (string_of_register r) ^ 
				       (if b=0 then "" else 
					  (if b > 0 then "+" else "")
				       ^ (string_of_int b)) ^ "]"
				      )
				      r'
			    )
			|  _ -> raise (Asm_error "trying to unspill a variable not on the stack and not in the same lexical scope")
		    else
		      (
			let old_reg = v.var_reg in
			  v.var_reg <- None;
			  let r = code_of_var v in
			    v.var_reg <- old_reg;
			    Instr_sparc.store r' ( r )
		  )
		)
	  | Spill(lex_scope, r, vd) ->
	      if lex_scope then
		match !(vd.vd_pos) with
		  | Pos_base b ->
		      (* using %g1 to load the display *)
		      get_display (get_vd_lex vd) (Reg_int 1) ^
		      Instr_sparc.store (string_of_register r) ("[" ^ (string_of_register (Reg_int 1)) ^ (if b = 0 then "" else (if b < 0 then "" else "+") ^ (string_of_int b)) ^ "]")
		  | _ -> raise (Asm_error "trying to spill a variable not on the stack and not in the same lexical scope")
	      else
		Instr_sparc.store (string_of_register r) (string_of_stack_pos fp (get_vd_pos vd)) 
	  | Call n -> Instr_sparc.call n
(*	  | Push_o(v,len) ->
	      let r = code_of_var v in
		Instr_sparc.sub "%sp" (string_of_int len) "%sp" ^
*)
	  | Set_var (f,v) ->
	      let l1 = get_fresh_label() in
	      let l2 = get_fresh_label() in
		Instr_sparc.branch (jump_of_flag f) l1 ^ Instr_sparc.nop ^
		Instr_sparc.move "%g0" (string_of_register (get_var_reg v)) ^
		Instr_sparc.jmp l2 ^ Instr_sparc.nop ^
		l1 ^ ":\n" ^
		Instr_sparc.move "1" (string_of_register (get_var_reg v)) ^
		l2 ^ ":\n"
	  | Return v ->
	      let r = code_of_var v in
		Instr_sparc.move r "%i0" ^
		Instr_sparc.restore ^ Instr_sparc.ret
	  | Nop -> Instr_sparc.nop
	  | Push v -> let r = code_of_var v
	    in Instr_sparc.push r
	  | Code c -> c
	  | Check_avail_with_flags _
	  | Check_avail _
	  | Assert_avail _
	  | Spill_var _
	  | Spill_all
	  | Read_flag _
	  | Proc _
	  | Fun _ -> raise (Asm_error "internal error in assembly generation")
	  | _ -> raise(Asm_error "instruction not implemented")
    in

    let rec code_of_base_block bb =
      (List.fold_left (fun s i -> s ^ code_of_instr i) "" bb.bb_block) ^ code_of_link bb.bb_link;

    and code_of_link = function
      | If(v, b1, b2, b3) ->
	  let c1 = code_of_base_block b1
	  and c2 = code_of_base_block b2
	  and c3 = code_of_base_block b3 in
	  let l2 = get_fresh_label () and l3 = get_fresh_label () in
	    (
	      match v.var_def.vd_value with
		| Some(Bool b) ->
		    (if b then c1 else c2) ^ c3
		| _ ->
		    (
		      match v.var_reg with
			| Some (Reg_flag f) ->
			    Instr_sparc.branch (jump_of_flag(neg_flag f)) l2 ^ 
			    Instr_sparc.nop
			| Some (Reg_int r) ->
			    Instr_sparc.cmp "%g0" 
			      (string_of_register (Reg_int r)) ^ 
			    Instr_sparc.branch "be" l2 ^ 
			    Instr_sparc.nop
			| None -> failwith "no value for jump's bool"
		    ) ^
		    c1 ^ "\tba\t" ^ l3 ^ "\n" ^
		    l2 ^ ":\n" ^ c2 ^
		    l3 ^ ":\n" ^ c3
	    )
      | While(v, be, b1, b2) ->
	  let ce = code_of_base_block be
	  and c1 = code_of_base_block b1
	  and c2 = code_of_base_block b2 in
	  let l1 = get_fresh_label () and l2 = get_fresh_label () in
	    (
	      match v.var_def.vd_value with
		| Some(Bool b) ->
		    if b then
		      l1 ^ ":\n" ^ c1 ^ "\tba\t" ^ l1 ^ "\n"
		    else
		      c2
		| _ ->
		    l1 ^ ":\n" ^ ce ^
		    (
		      match v.var_reg with
			| Some (Reg_flag f) ->
			    "\t" ^ (jump_of_flag (neg_flag f)) ^ "\t" ^ l2 ^ "\n"
			| Some (Reg_int r) ->
			    Instr_sparc.cmp "%g0" 
			      (string_of_register (Reg_int r)) ^
			    "\tbe\t" ^ l2 ^ "\n"
			| None -> failwith "no value for jump's bool"
		    ) ^
		    c1 ^ "\tba\t" ^ l1 ^ "\n" ^
		    l2 ^ ":\n" ^ c2
	    )
      | End -> ""
    in
      ret := !ret ^
      ".globl " ^ p.proc_name ^ "\n" ^
      "\t.type\t" ^ p.proc_name ^ ", @function\n" ^ p.proc_name ^ ":\n" ^
      (Instr_sparc.save "%sp" 
	 (string_of_int (-p.proc_stack_size - 104)) "%sp") ^
      Instr_sparc.set 
	       ("(DISPLAY+" ^ (string_of_int (4 * p.proc_lex_depth)) ^ ")")
	       "%l0" ^
      Instr_sparc.store "%fp" "[%l0]"; 
      (** Let's save the ARP in the display. *)
      ret := !ret ^ code_of_base_block p.proc_value;
      if not p.proc_is_fun then
	ret := !ret ^ Instr_sparc.move "%g0" "%i0" ^
	Instr_sparc.ret ^ Instr_sparc.restore 
  in

  (** Set the positions of arguments on the stack and in the registers. *)
  let set_args_pos args = 
    (* TODO: support for arrays and records *)
    let size = ref 8 in
    let argv = Array.of_list args in
      for i = 0 to min 5 (Array.length argv - 1) do
	argv.(i).vd_pos := Pos_reg(Reg_int (i+24));
      done;
      for i = 6 to Array.length argv - 1 do
	argv.(i).vd_pos := Pos_base (!size);
	size := !size + (sizeof(get_vd_type argv.(i)))
      done
  in

  (** Get the size needed by the procedure on the stack. This function also sets the positions of the variables on the stack ([vd_pos]). *)
  let rec get_stack_size pv =
    List.fold_left (fun size vd ->
		      let set_on_the_stack () =
			vd.vd_pos := Pos_base (-size - (sizeof (get_vd_type vd)));
			size + (sizeof (get_vd_type vd))
		      in
			match (get_vd_pos vd), get_vd_type vd with
			  | Pos_none, _ when (get_vd_ots vd) (* && not (is_constant (new_var v)) *) ->
			      set_on_the_stack ()
			  | _, Other _ ->
			      set_on_the_stack ()
			  | _ -> size
		   ) 0 pv
  in
    List.iter
      (fun p ->
	 set_args_pos p.proc_args;
	 allocate_registers p.proc_value p.proc_lex_depth;
(*	 if !Params.local_opts then
	 local_opt_base_block p.proc_value; *)
	 p.proc_stack_size <- get_stack_size p.proc_vars
      ) prog.prog_procs;
    ret := "\t.file\t\"" ^ prog.prog_source_name ^ "\"\n" ^
    "\t.section\t.data\nDISPLAY:\n\t.zero\t" ^ (string_of_int (4 * (prog.prog_max_depth + 1))) ^ "\n";
    ret := !ret ^ "\t.text\n";
    List.iter (fun p -> code_of_proc p) prog.prog_procs; !ret ^ "\t.size\tmain, .-main\n" ^ "\t.section\t.note.GNU-stack,\"\",@progbits\n" ^ "\t.ident\t\"Bifton 0.1.0\"\n"
